from collections import OrderedDict

from ovos_number_parser.util import (convert_to_mixed_fraction)

_NUM_STRING_SL = {
    0: 'nič',
    1: 'ena',
    2: 'dve',
    3: 'tri',
    4: 'štiri',
    5: 'pet',
    6: 'šest',
    7: 'sedem',
    8: 'osem',
    9: 'devet',
    10: 'deset',
    11: 'enajst',
    12: 'dvanajst',
    13: 'trinajst',
    14: 'štirinajst',
    15: 'petnajst',
    16: 'šestnajst',
    17: 'sedemnajst',
    18: 'osemnajst',
    19: 'devetnajst',
    20: 'dvajset',
    30: 'trideset',
    40: 'štirideset',
    50: 'petdeset',
    60: 'šestdeset',
    70: 'sedemdeset',
    80: 'osemdeset',
    90: 'devetdeset'
}

_FRACTION_STRING_SL = {
    2: 'polovica',
    3: 'tretjina',
    4: 'četrtina',
    5: 'petina',
    6: 'šestina',
    7: 'sedmina',
    8: 'osmina',
    9: 'devetina',
    10: 'desetina',
    11: 'enajstina',
    12: 'dvanajstina',
    13: 'trinajstina',
    14: 'štirinajstina',
    15: 'petnajstina',
    16: 'šestnajstina',
    17: 'sedemnajstina',
    18: 'osemnajstina',
    19: 'devetnajstina',
    20: 'dvajsetina'
}

_LONG_SCALE_SL = OrderedDict([
    (100, 'sto'),
    (1000, 'tisoč'),
    (1000000, 'milijon'),
    (1e12, 'bilijon'),
    (1e18, 'trilijon'),
    (1e24, 'kvadrilijon'),
    (1e30, 'kvintilijon'),
    (1e36, 'sekstilijon'),
    (1e42, 'septilijon'),
    (1e48, 'oktilijon'),
    (1e54, 'nonilijon'),
    (1e60, 'decilijon')
    # TODO > 1e63
])

_SHORT_SCALE_SL = OrderedDict([
    (100, 'sto'),
    (1000, 'tisoč'),
    (1000000, 'milijon'),
    (1e9, 'bilijon'),
    (1e12, 'trilijon'),
    (1e15, 'kvadrilijon'),
    (1e18, 'kvintilijon'),
    (1e21, 'sekstilijon'),
    (1e24, 'septilijon'),
    (1e27, 'oktilijon'),
    (1e30, 'nonilijon'),
    (1e33, 'decilijon')
    # TODO > 1e33
])

_ORDINAL_BASE_SL = {
    1: 'prvi',
    2: 'drugi',
    3: 'tretji',
    4: 'četrti',
    5: 'peti',
    6: 'šesti',
    7: 'sedmi',
    8: 'osmi',
    9: 'deveti',
    10: 'deseti',
    11: 'enajsti',
    12: 'dvanajsti',
    13: 'trinajsti',
    14: 'štirinajsti',
    15: 'petnajsti',
    16: 'šestnajsti',
    17: 'sedemnajsti',
    18: 'osemnajsti',
    19: 'devetnajsti',
    20: 'dvajseti',
    30: 'trideseti',
    40: 'štirideseti',
    50: 'petdeseti',
    60: 'šestdeseti',
    70: 'sedemdeseti',
    80: 'osemdeseti',
    90: 'devetdeseti',
    1e2: 'stoti',
    1e3: 'tisoči'
}

_LONG_ORDINAL_SL = {
    1e6: 'milijonti',
    1e12: 'bilijonti',
    1e18: 'trilijonti',
    1e24: 'kvadrilijonti',
    1e30: 'kvintiljonti',
    1e36: 'sekstilijonti',
    1e42: 'septilijonti',
    1e48: 'oktilijonti',
    1e54: 'nonilijonti',
    1e60: 'decilijonti'
    # TODO > 1e60
}
_LONG_ORDINAL_SL.update(_ORDINAL_BASE_SL)

_SHORT_ORDINAL_SL = {
    1e6: 'milijonti',
    1e9: 'bilijonti',
    1e12: 'trilijonti',
    1e15: 'kvadrilijonti',
    1e18: 'kvintiljonti',
    1e21: 'sekstilijonti',
    1e24: 'septilijonti',
    1e27: 'oktilijonti',
    1e30: 'nonilijonti',
    1e33: 'decilijonti'
    # TODO > 1e33
}
_SHORT_ORDINAL_SL.update(_ORDINAL_BASE_SL)


def nice_number_sl(number, speech=True, denominators=None):
    """ Slovenian helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "2 in polovica" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    if denominators is None:
        denominators = range(1, 21)
    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_SL[den]
    if whole == 0:
        return_string = '{} {}'.format(num, den_str)
    else:
        return_string = '{} in {} {}'.format(whole, num, den_str)

    if num % 100 == 1:
        pass
    elif num % 100 == 2:
        return_string = return_string[:-1] + 'i'
    elif num % 100 == 3 or num % 100 == 4:
        return_string = return_string[:-1] + 'e'
    else:
        return_string = return_string[:-1]

    return return_string


def pronounce_number_sl(num, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'pet celih dve'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    # deal with infinity
    if num == float("inf"):
        return "neskončno"
    elif num == float("-inf"):
        return "minus neskončno"
    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            if ordinals:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} krat deset na {}{}'.format(
                    'minus ' if float(n) < 0 else '',
                    pronounce_number_sl(
                        abs(float(n)), places, short_scale, False, ordinals=False),
                    'minus ' if power < 0 else '',
                    pronounce_number_sl(abs(power), places, short_scale, False, ordinals=True))
            else:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} krat deset na {}{}'.format(
                    'minus ' if float(n) < 0 else '',
                    pronounce_number_sl(
                        abs(float(n)), places, short_scale, False),
                    'minus ' if power < 0 else '',
                    pronounce_number_sl(abs(power), places, short_scale, False))

    if short_scale:
        number_names = _NUM_STRING_SL.copy()
        number_names.update(_SHORT_SCALE_SL)
    else:
        number_names = _NUM_STRING_SL.copy()
        number_names.update(_LONG_SCALE_SL)

    digits = [number_names[n] for n in range(0, 20)]

    tens = [number_names[n] for n in range(10, 100, 10)]

    if short_scale:
        hundreds = [_SHORT_SCALE_SL[n] for n in _SHORT_SCALE_SL.keys()]
    else:
        hundreds = [_LONG_SCALE_SL[n] for n in _LONG_SCALE_SL.keys()]

    # deal with negatives
    result = ""
    if num < 0:
        result = "minus "
    num = abs(num)

    # check for a direct match
    if num in number_names and not ordinals:
        result += number_names[num]
    else:
        def _sub_thousand(n, ordinals=False, is_male=False):
            assert 0 <= n <= 999
            if n in _SHORT_ORDINAL_SL and ordinals:
                return _SHORT_ORDINAL_SL[n]
            if n <= 19:
                if is_male and n == 2:
                    return digits[n][:-1] + "a"
                return digits[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                sub = _sub_thousand(r, False)
                if r == 2:
                    sub = sub[:-1] + "a"
                return ((sub + "in") if r else "") + (
                    tens[q - 1]) + ("i" if ordinals else "")
            else:
                q, r = divmod(n, 100)
                if q == 1:
                    qstr = ""
                else:
                    qstr = digits[q]
                return (qstr + "sto" + (
                    " " + _sub_thousand(r, ordinals) if r else ""))

        def _plural_hundreds(n, hundred, ordi=True):
            if hundred[-3:] != "jon":
                if ordi:
                    return hundred + "i"

                return hundred

            if n < 1000 or short_scale:
                if ordi:
                    return hundred + "ti"

                if n % 100 == 1:
                    return hundred
                elif n % 100 == 2:
                    return hundred + "a"
                elif n % 100 == 3 or n % 100 == 4:
                    return hundred + "i"
                else:
                    return hundred + "ov"
            else:
                n //= 1000

                if ordi:
                    return hundred[:-3] + "jardti"

                if n % 100 == 1:
                    return hundred[:-3] + "jarda"
                elif n % 100 == 2:
                    return hundred[:-3] + "jardi"
                elif n % 100 == 3 or n % 100 == 4:
                    return hundred[:-3] + "jarde"
                else:
                    return hundred[:-3] + "jard"

        def _short_scale(n):
            if n >= max(_SHORT_SCALE_SL.keys()):
                return "neskončno"
            ordi = ordinals

            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []

            split = _split_by(n, 1000)
            if ordinals and len([a for a in split if a > 0]) == 1:
                ordi_force = True
            else:
                ordi_force = False

            for i, z in enumerate(split):
                if not z:
                    continue

                if z == 1 and i == 1:
                    number = ""
                elif z > 100 and z % 100 == 2:
                    number = _sub_thousand(z, not i and ordi, is_male=True)
                elif z > 100 and z % 100 == 3:
                    number = _sub_thousand(z, not i and ordi) + "je"
                elif z > 1 or i == 0 or ordi:
                    number = _sub_thousand(z, not i and ordi)
                else:
                    number = ""

                if i:
                    if i >= len(hundreds):
                        return ""
                    if z > 1:
                        number += " "
                    number += _plural_hundreds(
                        z, hundreds[i], True if ordi_force else not i and ordi)
                res.append(number)
                ordi = False

            return " ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        def _long_scale(n):
            if n >= max(_LONG_SCALE_SL.keys()):
                return "neskončno"
            ordi = ordinals
            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []

            split = _split_by(n, 1000000)
            if ordinals and len([a for a in split if a > 0]) == 1:
                ordi_force = True
            else:
                ordi_force = False

            for i, z in enumerate(split):
                if not z:
                    continue

                number = pronounce_number_sl(z, places, True, scientific)
                if z > 100:
                    add = number.split()[0] + " "
                else:
                    add = ""
                if z % 100 == 2 and i >= 1:
                    number = add + digits[2][:-1] + "a"
                if z % 100 == 3 and i >= 1:
                    number = add + digits[3] + "je"

                # strip off the comma after the thousand
                if i:
                    if i >= len(hundreds):
                        return ""
                    # plus one as we skip 'thousand'
                    # (and 'hundred', but this is excluded by index value)
                    hundred = _plural_hundreds(
                        z, hundreds[i + 1], True if ordi_force else ordi and not i)

                    if z >= 1000:
                        z //= 1000
                        number = pronounce_number_sl(z, places, True, scientific,
                                                     ordinals=True if ordi_force else ordi and not i)

                    if z == 1:
                        number = hundred
                    else:
                        number += " " + hundred
                res.append(number)
            return " ".join(reversed(res))

        if short_scale:
            result += _short_scale(num)
        else:
            result += _long_scale(num)

    if ordinals:
        result = result.replace(" ", "")

    # deal with scientific notation unpronounceable as number
    if (not result or result == "neskončno") and "e" in str(num):
        return pronounce_number_sl(num, places, short_scale, scientific=True)
    # Deal with fractional part
    elif not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "minus " or not result):
            result += "nič"

        if int(abs(num)) % 100 == 1:
            result += " cela"
        elif int(abs(num)) % 100 == 2:
            result += " celi"
        elif int(abs(num)) % 100 == 3 or int(abs(num)) % 100 == 4:
            result += " cele"
        else:
            result += " celih"

        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result



#
# Copyright 2021 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict

from ovos_number_parser.util import (invert_dict, convert_to_mixed_fraction, tokenize, look_for_fractions,
                                     partition_list, is_numeric, Token, ReplaceableNumber)

_NUM_STRING_AZ = {
    0: 'sıfır',
    1: 'bir',
    2: 'iki',
    3: 'üç',
    4: 'dörd',
    5: 'beş',
    6: 'altı',
    7: 'yeddi',
    8: 'səkkiz',
    9: 'doqquz',
    10: 'on',
    11: 'on bir',
    12: 'on iki',
    13: 'on üç',
    14: 'on dörd',
    15: 'on beş',
    16: 'on altı',
    17: 'on yeddi',
    18: 'on səkkiz',
    19: 'on doqquz',
    20: 'iyirmi',
    30: 'otuz',
    40: 'qırx',
    50: 'əlli',
    60: 'altmış',
    70: 'yetmiş',
    80: 'səksən',
    90: 'doxsan'
}

_FRACTION_STRING_AZ = {
    2: 'ikidə',
    3: 'üçdə',
    4: 'dörddə',
    5: 'beşdə',
    6: 'altıda',
    7: 'yeddidə',
    8: 'səkkizdə',
    9: 'doqquzda',
    10: 'onda',
    11: 'on birdə',
    12: 'on ikidə',
    13: 'on üçdə',
    14: 'on dörddə',
    15: 'on beşdə',
    16: 'on altıda',
    17: 'on yeddidə',
    18: 'on səkkizdə',
    19: 'on doqquzda',
    20: 'iyirmidə',
    30: 'otuzda',
    40: 'qırxda',
    50: 'əllidə',
    60: 'altmışda',
    70: 'yetmişdə',
    80: 'səksəndə',
    90: 'doxsanda',
    1e2: 'yüzdə',
    1e3: 'mində'
}

_LONG_SCALE_AZ = OrderedDict([
    (100, 'yüz'),
    (1000, 'min'),
    (1000000, 'milyon'),
    (1e12, "milyard"),
    (1e18, 'trilyon'),
    (1e24, "kvadrilyon"),
    (1e30, "kvintilyon"),
    (1e36, "sekstilyon"),
    (1e42, "septilyon"),
    (1e48, "oktilyon"),
    (1e54, "nonilyon"),
    (1e60, "dekilyon")
])

_SHORT_SCALE_AZ = OrderedDict([
    (100, 'yüz'),
    (1000, 'min'),
    (1000000, 'milyon'),
    (1e9, "milyard"),
    (1e12, 'trilyon'),
    (1e15, "kvadrilyon"),
    (1e18, "kvintilyon"),
    (1e21, "sekstilyon"),
    (1e24, "septilyon"),
    (1e27, "oktilyon"),
    (1e30, "nonilyon"),
    (1e33, "dekilyon")
])

_ORDINAL_BASE_AZ = {
    1: 'birinci',
    2: 'ikinci',
    3: 'üçüncü',
    4: 'dördüncü',
    5: 'beşinci',
    6: 'altıncı',
    7: 'yeddinci',
    8: 'səkkizinci',
    9: 'doqquzuncu',
    10: 'onuncu',
    11: 'on birinci',
    12: 'on ikinci',
    13: 'on üçüncü',
    14: 'on dördüncü',
    15: 'on beşinci',
    16: 'on altıncı',
    17: 'on yeddinci',
    18: 'on səkkizinci',
    19: 'on doqquzuncu',
    20: 'iyirminci',
    30: 'otuzuncu',
    40: "qırxıncı",
    50: "əllinci",
    60: "altmışıncı",
    70: "yetmışinci",
    80: "səksəninci",
    90: "doxsanınçı",
    1e2: "yüzüncü",
    1e3: "mininci"
}

_SHORT_ORDINAL_AZ = {
    1e6: "milyonuncu",
    1e9: "milyardıncı",
    1e12: "trilyonuncu",
    1e15: "kvadrilyonuncu",
    1e18: "kvintilyonuncu",
    1e21: "sekstilyonuncu",
    1e24: "septilyonuncu",
    1e27: "oktilyonuncu",
    1e30: "nonilyonuncu",
    1e33: "dekilyonuncu"
    # TODO > 1e-33
}
_SHORT_ORDINAL_AZ.update(_ORDINAL_BASE_AZ)

_LONG_ORDINAL_AZ = {
    1e6: "milyonuncu",
    1e12: "milyardıncı",
    1e18: "trilyonuncu",
    1e24: "kvadrilyonuncu",
    1e30: "kvintilyonuncu",
    1e36: "sekstilyonuncu",
    1e42: "septilyonuncu",
    1e48: "oktilyonuncu",
    1e54: "nonilyonuncu",
    1e60: "dekilyonuncu"
    # TODO > 1e60
}
_LONG_ORDINAL_AZ.update(_ORDINAL_BASE_AZ)

# negate next number (-2 = 0 - 2)
_NEGATIVES_AZ = {"mənfi", "minus"}

# sum the next number (iyirmi iki = 20 + 2)
_SUMS_AZ = {'on', '10', 'iyirmi', '20', 'otuz', '30', 'qırx', '40', 'əlli', '50',
            'altmış', '60', 'yetmiş', '70', 'səksən', '80', 'doxsan', '90'}

_HARD_VOWELS = ['a', 'ı', 'o', 'u']
_SOFT_VOWELS = ['e', 'ə', 'i', 'ö', 'ü']
_VOWELS = _HARD_VOWELS + _SOFT_VOWELS


def _get_last_vowel(word):
    is_last = True
    for char in word[::-1]:
        if char in _VOWELS:
            return char, is_last
        is_last = False

    return "", is_last


def _last_vowel_type(word):
    return _get_last_vowel(word)[0] in _HARD_VOWELS


def _get_ordinal_ak(word):
    last_vowel, is_last = _get_last_vowel(word)
    if not last_vowel:
        return ""

    if last_vowel in ["a", "ı"]:
        if is_last:
            return "ncı"
        return "ıncı"

    if last_vowel == ["e", "ə", "i"]:
        if is_last:
            return "nci"
        return "inci"

    if last_vowel in ["o", "u"]:
        if is_last:
            return "ncu"
        return "uncu"

    if last_vowel == ["ö", "ü"]:
        if is_last:
            return "ncü"
        return "üncü"


def _get_full_time_ak(hour):
    if hour in [1, 3, 4, 5, 8, 11]:
        return "ə"
    if hour in [2, 7, 12]:
        return "yə"
    if hour in [9, 10]:
        return "a"
    return "ya"


def _get_half_time_ak(hour):
    if hour in [1, 5, 8, 11]:
        return "in"
    if hour in [2, 7, 12]:
        return "nin"
    if hour in [3, 4]:
        return "ün"
    if hour in [9, 10]:
        return "un"
    return "nın"


def _get_daytime(hour):
    if hour < 6:
        return "gecə"
    if hour < 12:
        return "səhər"
    if hour < 18:
        return "gündüz"
    return "axşam"


def _generate_plurals_az(originals):
    """
    Return a new set or dict containing the plural form of the original values,

    In Azerbaijani this means appending 'lar' or 'lər' to them according to the last vowel in word.

    Args:
        originals set(str) or dict(str, any): values to pluralize

    Returns:
        set(str) or dict(str, any)

    """

    if isinstance(originals, dict):
        return {key + ('lar' if _last_vowel_type(key) else 'lər'): value for key, value in originals.items()}
    return {value + ('lar' if _last_vowel_type(value) else 'lər') for value in originals}


_MULTIPLIES_LONG_SCALE_AZ = set(_LONG_SCALE_AZ.values()) | \
                            set(_LONG_SCALE_AZ.values())

_MULTIPLIES_SHORT_SCALE_AZ = set(_SHORT_SCALE_AZ.values()) | \
                             set(_SHORT_SCALE_AZ.values())

# split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
_FRACTION_MARKER_AZ = {"və"}

# decimal marker ( 1 nöqtə 5 = 1 + 0.5)
_DECIMAL_MARKER_AZ = {"nöqtə"}

_STRING_NUM_AZ = invert_dict(_NUM_STRING_AZ)

_SPOKEN_EXTRA_NUM_AZ = {
    "yarım": 0.5,
    "üçdəbir": 1 / 3,
    "dörddəbir": 1 / 4
}

_STRING_SHORT_ORDINAL_AZ = invert_dict(_SHORT_ORDINAL_AZ)
_STRING_LONG_ORDINAL_AZ = invert_dict(_LONG_ORDINAL_AZ)


def nice_number_az(number, speech=True, denominators=range(1, 21)):
    """ Azerbaijani helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 yarım" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_AZ[den]
    if whole == 0:
        if den == 2:
            return 'yarım'
        return '{} {}'.format(den_str, num)
    if den == 2:
        return '{} yarım'.format(whole)
    return '{} və {} {}'.format(whole, den_str, num)


def pronounce_number_az(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'beş nöqtə iki'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    num = number
    # deal with infinity
    if num == float("inf"):
        return "sonsuzluq"
    elif num == float("-inf"):
        return "mənfi sonsuzluq"
    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            if ordinals:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} vurulsun on üstü {}{}'.format(
                    'mənfi ' if float(n) < 0 else '',
                    pronounce_number_az(
                        abs(float(n)), places, short_scale, False, ordinals=False),
                    'mənfi ' if power < 0 else '',
                    pronounce_number_az(abs(power), places, short_scale, False, ordinals=True))
            else:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} vurulsun on üstü {}{}'.format(
                    'mənfi ' if float(n) < 0 else '',
                    pronounce_number_az(
                        abs(float(n)), places, short_scale, False),
                    'mənfi ' if power < 0 else '',
                    pronounce_number_az(abs(power), places, short_scale, False))

    if short_scale:
        number_names = _NUM_STRING_AZ.copy()
        number_names.update(_SHORT_SCALE_AZ)
    else:
        number_names = _NUM_STRING_AZ.copy()
        number_names.update(_LONG_SCALE_AZ)

    digits = [number_names[n] for n in range(0, 20)]

    tens = [number_names[n] for n in range(10, 100, 10)]

    if short_scale:
        hundreds = [_SHORT_SCALE_AZ[n] for n in _SHORT_SCALE_AZ.keys()]
    else:
        hundreds = [_LONG_SCALE_AZ[n] for n in _LONG_SCALE_AZ.keys()]

    # deal with negatives
    result = ""
    if num < 0:
        # result = "mənfi " if scientific else "minus "
        result = "mənfi "
    num = abs(num)

    # check for a direct match
    if num in number_names and not ordinals:
        if num > 1000:
            result += "bir "
        result += number_names[num]
    else:
        def _sub_thousand(n, ordinals=False):
            assert 0 <= n <= 999
            if n in _SHORT_ORDINAL_AZ and ordinals:
                return _SHORT_ORDINAL_AZ[n]
            if n <= 19:
                return digits[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                return tens[q - 1] + (" " + _sub_thousand(r, ordinals) if r
                                      else "")
            else:
                q, r = divmod(n, 100)
                return (digits[q] + " " if q != 1 else "") + "yüz" + (
                    " " + _sub_thousand(r, ordinals) if r else "")

        def _short_scale(n):
            if n >= 999 * max(_SHORT_SCALE_AZ.keys()):
                return "sonsuzluq"
            ordi = ordinals

            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000)):
                if not z:
                    continue

                number = _sub_thousand(z, not i and ordi)

                if i:
                    if i >= len(hundreds):
                        return ""
                    number += " "
                    if ordi:

                        if i * 1000 in _SHORT_ORDINAL_AZ:
                            if z == 1:
                                number = _SHORT_ORDINAL_AZ[i * 1000]
                            else:
                                number += _SHORT_ORDINAL_AZ[i * 1000]
                        else:
                            if n not in _SHORT_SCALE_AZ:
                                num = int("1" + "0" * (len(str(n)) - 2))

                                number += _SHORT_SCALE_AZ[num] + _get_ordinal_ak(_SHORT_SCALE_AZ[num])
                            else:
                                number = _SHORT_SCALE_AZ[n] + _get_ordinal_ak(_SHORT_SCALE_AZ[n])
                    else:
                        number += hundreds[i]
                if number.startswith("bir min"):
                    number = number[4:]
                res.append(number)
                ordi = False

            return ", ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        def _long_scale(n):
            if n >= max(_LONG_SCALE_AZ.keys()):
                return "sonsuzluq"
            ordi = ordinals
            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000000)):
                if not z:
                    continue
                number = pronounce_number_az(z, places, True, scientific,
                                             ordinals=ordi and not i)
                # strip off the comma after the thousand
                if i:
                    if i >= len(hundreds):
                        return ""
                    # plus one as we skip 'thousand'
                    # (and 'hundred', but this is excluded by index value)
                    number = number.replace(',', '')

                    if ordi:
                        if i * 1000000 in _LONG_ORDINAL_AZ:
                            if z == 1:
                                number = _LONG_ORDINAL_AZ[
                                    (i + 1) * 1000000]
                            else:
                                number += _LONG_ORDINAL_AZ[
                                    (i + 1) * 1000000]
                        else:
                            if n not in _LONG_SCALE_AZ:
                                num = int("1" + "0" * (len(str(n)) - 2))

                                number += " " + _LONG_SCALE_AZ[
                                    num] + _get_ordinal_ak(_LONG_SCALE_AZ[num])
                            else:
                                number = " " + _LONG_SCALE_AZ[n] + _get_ordinal_ak(_LONG_SCALE_AZ[n])
                    else:

                        number += " " + hundreds[i + 1]
                res.append(number)
            return ", ".join(reversed(res))

        if short_scale:
            result += _short_scale(num)
        else:
            result += _long_scale(num)

    # deal with scientific notation unpronounceable as number
    if not result and "e" in str(num):
        return pronounce_number_az(num, places, short_scale, scientific=True)
    # Deal with fractional part
    elif not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "mənfi " or not result):
            result += "sıfır"
        result += " nöqtə"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result


def numbers_to_digits_az(text, short_scale=True, ordinals=False):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numbers should be used.
        ordinals boolean: True if ordinals (e.g. birinci, ikinci, üçüncü) should
                          be parsed to their number values (1, 2, 3...)

    Returns:
        str
        The original text, with numbers subbed in where appropriate.

    """
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_az(tokens, short_scale, ordinals)

    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_az(tokens, short_scale=True,
                                  ordinals=False, fractional_numbers=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (birinci, ikinci, üçüncü, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_az(tokens, short_scale,
                                         ordinals, fractional_numbers)
        if not to_replace:
            break

        results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_az(tokens, short_scale=True,
                                 ordinals=False, fractional_numbers=True):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = \
        _extract_number_with_text_az_helper(tokens, short_scale,
                                            ordinals, fractional_numbers)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_az_helper(tokens,
                                        short_scale=True, ordinals=False,
                                        fractional_numbers=True):
    """
    Helper for _extract_number_with_text_az.

    This contains the real logic for parsing, but produces
    a result that needs a little cleaning (specific, it may
    contain leading articles that can be trimmed off).

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:

    Returns:
        int or float, [Tokens]

    """
    if fractional_numbers:
        fraction, fraction_text = \
            _extract_fraction_with_text_az(tokens, short_scale, ordinals)
        if fraction:
            # print("fraction")
            return fraction, fraction_text

        decimal, decimal_text = \
            _extract_decimal_with_text_az(tokens, short_scale, ordinals)
        if decimal:
            # print("decimal")
            return decimal, decimal_text

    return _extract_whole_number_with_text_az(tokens, short_scale, ordinals)


def _extract_fraction_with_text_az(tokens, short_scale, ordinals):
    """
    Extract fraction numbers from a string.

    This function handles text such as '2 və dörddə üç'. Note that "yarım" or
    similar will be parsed by the whole number function.

    Args:
        tokens [Token]: words and their indexes in the original string.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (int or float, [Token])
        The value found, and the list of relevant tokens.
        (None, None) if no fraction value is found.

    """
    for c in _FRACTION_MARKER_AZ:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_az(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_az(partitions[2], short_scale,
                                              ordinals, fractional_numbers=True)

            if not numbers1 or not numbers2:
                return None, None

            # ensure first is not a fraction and second is a fraction
            num1 = numbers1[-1]
            num2 = numbers2[0]
            if num1.value >= 1 and 0 < num2.value < 1:
                return num1.value + num2.value, \
                       num1.tokens + partitions[1] + num2.tokens

    return None, None


def _extract_decimal_with_text_az(tokens, short_scale, ordinals):
    """
    Extract decimal numbers from a string.

    This function handles text such as '2 nöqtə 5'.

    Notes:
        While this is a helper for extractnumber_az, it also depends on
        extractnumber_az, to parse out the components of the decimal.

        This does not currently handle things like:
            number dot number number number

    Args:
        tokens [Token]: The text to parse.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (float, [Token])
        The value found and relevant tokens.
        (None, None) if no decimal value is found.

    """
    for c in _DECIMAL_MARKER_AZ:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_az(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_az(partitions[2], short_scale,
                                              ordinals, fractional_numbers=False)
            if not numbers1 or not numbers2:
                return None, None

            number = numbers1[-1]
            decimal = numbers2[0]

            # TODO handle number dot number number number
            if "." not in str(decimal.text):
                return number.value + float('0.' + str(decimal.value)), \
                       number.tokens + partitions[1] + decimal.tokens
    return None, None


def _extract_whole_number_with_text_az(tokens, short_scale, ordinals):
    """
    Handle numbers not handled by the decimal or fraction functions. This is
    generally whole numbers. Note that phrases such as "yarım" will be
    handled by this function.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:

    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.

    """
    multiplies, string_num_ordinal, string_num_scale = \
        _initialize_number_data_az(short_scale, speech=ordinals is not None)

    number_words = []  # type: List[Token]
    val = False
    prev_val = None
    next_val = None
    to_sum = []
    # print(tokens, ordinals)
    for idx, token in enumerate(tokens):
        current_val = None
        if next_val:
            next_val = None
            continue

        word = token.word.lower()
        if word in _NEGATIVES_AZ:
            number_words.append(token)
            continue

        prev_word = tokens[idx - 1].word.lower() if idx > 0 else ""
        next_word = tokens[idx + 1].word.lower() if idx + 1 < len(tokens) else ""
        # print(prev_word, word, next_word, number_words)
        if word not in string_num_scale and \
                word not in _STRING_NUM_AZ and \
                word not in _SUMS_AZ and \
                word not in multiplies and \
                not (ordinals and word in string_num_ordinal) and \
                not is_numeric(word) and \
                not is_fractional_az(word, short_scale=short_scale) and \
                not look_for_fractions(word.split('/')):
            # print("a1")
            words_only = [token.word for token in number_words]

            if number_words and not all([w.lower() in
                                         _NEGATIVES_AZ for w in words_only]):
                break
            else:
                number_words = []
                continue
        elif word not in multiplies \
                and word not in _SPOKEN_EXTRA_NUM_AZ \
                and prev_word not in multiplies \
                and prev_word not in _SUMS_AZ \
                and not (ordinals and prev_word in string_num_ordinal) \
                and prev_word not in _NEGATIVES_AZ:
            number_words = [token]
            # print("a2")
        elif prev_word in _SUMS_AZ and word in _SUMS_AZ:
            number_words = [token]
            # print("a3")
        elif ordinals is None and \
                (word in string_num_ordinal or word in _SPOKEN_EXTRA_NUM_AZ):
            # print("a4")
            # flagged to ignore this token
            continue
        else:
            # print("a5")
            number_words.append(token)

        # is this word already a number ?
        if is_numeric(word):
            # print("b")
            if word.isdigit():  # doesn't work with decimals
                val = int(word)
            else:
                val = float(word)
            current_val = val

        # is this word the name of a number ?
        if word in _STRING_NUM_AZ:
            val = _STRING_NUM_AZ.get(word)
            current_val = val
            # print("c1", current_val)
        elif word in string_num_scale:
            val = string_num_scale.get(word)
            current_val = val
            # print("c2")
        elif ordinals and word in string_num_ordinal:
            val = string_num_ordinal[word]
            current_val = val
            # print("c3")
        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if (prev_word in _SUMS_AZ and val and val < 10) or all([prev_word in
                                                                multiplies,
                                                                val < prev_val if prev_val else False]):
            val = prev_val + val
            # print("d")

        # is the prev word a number and should we multiply it?
        # twenty hundred, six hundred
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val
            # print("e")

        # is this a spoken fraction?
        # 1 yarım fincan - yarım fincan
        if current_val is None and not (ordinals is None and word in _SPOKEN_EXTRA_NUM_AZ):
            val = is_fractional_az(word, short_scale=short_scale,
                                   spoken=ordinals is not None)
            if val:
                if prev_val:
                    val += prev_val
                current_val = val
                # print("f", current_val, prev_val)
                if word in _SPOKEN_EXTRA_NUM_AZ:
                    break

        # dörddə bir
        if ordinals is False:
            temp = prev_val
            prev_val = is_fractional_az(prev_word, short_scale=short_scale)
            if prev_val:
                if not val:
                    val = 1
                val = val * prev_val
                if idx + 1 < len(tokens):
                    number_words.append(tokens[idx + 1])
            else:
                prev_val = temp
            # print("g", prev_val)

        # is this a negative number?
        if val and prev_word and prev_word in _NEGATIVES_AZ:
            val = 0 - val
            # print("h")

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])
                current_val = val
            # print("i")

        else:
            if current_val and all([
                prev_word in _SUMS_AZ,
                word not in _SUMS_AZ,
                word not in multiplies,
                current_val >= 10]):
                # Backtrack - we've got numbers we can't sum.
                # print("j", number_words, prev_val)
                number_words.pop()
                val = prev_val
                break
            prev_val = val

            if word in multiplies and next_word not in multiplies:
                # handle long numbers
                # six hundred sixty six
                # two million five hundred thousand
                #
                # This logic is somewhat complex, and warrants
                # extensive documentation for the next coder's sake.
                #
                # The current word is a power of ten. `current_val` is
                # its integer value. `val` is our working sum
                # (above, when `current_val` is 1 million, `val` is
                # 2 million.)
                #
                # We have a dict `string_num_scale` containing [value, word]
                # pairs for "all" powers of ten: string_num_scale[10] == "ten.
                #
                # We need go over the rest of the tokens, looking for other
                # powers of ten. If we find one, we compare it with the current
                # value, to see if it's smaller than the current power of ten.
                #
                # Numbers which are not powers of ten will be passed over.
                #
                # If all the remaining powers of ten are smaller than our
                # current value, we can set the current value aside for later,
                # and begin extracting another portion of our final result.
                # For example, suppose we have the following string.
                # The current word is "million".`val` is 9000000.
                # `current_val` is 1000000.
                #
                #    "nine **million** nine *hundred* seven **thousand**
                #     six *hundred* fifty seven"
                #
                # Iterating over the rest of the string, the current
                # value is larger than all remaining powers of ten.
                #
                # The if statement passes, and nine million (9000000)
                # is appended to `to_sum`.
                #
                # The main variables are reset, and the main loop begins
                # assembling another number, which will also be appended
                # under the same conditions.
                #
                # By the end of the main loop, to_sum will be a list of each
                # "place" from 100 up: [9000000, 907000, 600]
                #
                # The final three digits will be added to the sum of that list
                # at the end of the main loop, to produce the extracted number:
                #
                #    sum([9000000, 907000, 600]) + 57
                # == 9,000,000 + 907,000 + 600 + 57
                # == 9,907,657
                #
                # >>> foo = "nine million nine hundred seven thousand six
                #            hundred fifty seven"
                # >>> extract_number(foo)
                # 9907657
                # print("k", tokens[idx+1:])
                time_to_sum = True
                for other_token in tokens[idx + 1:]:
                    if other_token.word.lower() in multiplies:
                        if string_num_scale[other_token.word.lower()] >= current_val:
                            time_to_sum = False
                        else:
                            continue
                    if not time_to_sum:
                        break
                if time_to_sum:
                    # print("l")
                    to_sum.append(val)
                    val = 0
                    prev_val = 0

    if val is not None and to_sum:
        # print("m", to_sum)
        val += sum(to_sum)
    # print(val, number_words, "end")
    return val, number_words


def _initialize_number_data_az(short_scale, speech=True):
    """
    Generate dictionaries of words to numbers, based on scale.

    This is a helper function for _extract_whole_number.

    Args:
        short_scale (bool):
        speech (bool): consider extra words (_SPOKEN_EXTRA_NUM_AZ) to be numbers

    Returns:
        (set(str), dict(str, number), dict(str, number))
        multiplies, string_num_ordinal, string_num_scale

    """
    multiplies = _MULTIPLIES_SHORT_SCALE_AZ if short_scale \
        else _MULTIPLIES_LONG_SCALE_AZ

    string_num_ordinal_az = _STRING_SHORT_ORDINAL_AZ if short_scale \
        else _STRING_LONG_ORDINAL_AZ

    string_num_scale_az = _SHORT_SCALE_AZ if short_scale else _LONG_SCALE_AZ
    string_num_scale_az = invert_dict(string_num_scale_az)

    return multiplies, string_num_ordinal_az, string_num_scale_az


def is_fractional_az(input_str, short_scale=True, spoken=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
        spoken (bool):
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """

    fracts = {"dörddəbir": 4, "yarım": 2, "üçdəbir": 3}
    for num in _FRACTION_STRING_AZ:
        if num > 2:
            fracts[_FRACTION_STRING_AZ[num]] = num

    if input_str.lower() in fracts and spoken:
        return 1.0 / fracts[input_str.lower()]
    return False


def extract_number_az(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    return _extract_number_with_text_az(tokenize(text.lower()),
                                        short_scale, ordinals).value



import math
from enum import IntEnum

from ovos_number_parser.util import (convert_to_mixed_fraction)

_FRACTION_STRING_FA = {
    2: 'دوم',
    3: 'سوم',
    4: 'چهارم',
    5: 'پنجم',
    6: 'ششم',
    7: 'هفتم',
    8: 'هشتم',
    9: 'نهم',
    10: 'دهم',
    11: 'یازدهم',
    12: 'دوازدهم',
    13: 'سیزدهم',
    14: 'چهاردهم',
    15: 'پونزدهم',
    16: 'شونزدهم',
    17: 'هیفدهم',
    18: 'هیجدهم',
    19: 'نوزدهم',
    20: 'بیستم'
}

_FARSI_ONES = [
    "",
    "یک",
    "دو",
    "سه",
    "چهار",
    "پنج",
    "شش",
    "هفت",
    "هشت",
    "نه",
    "ده",
    "یازده",
    "دوازده",
    "سیزده",
    "چهارده",
    "پونزده",
    "شونزده",
    "هیفده",
    "هیجده",
    "نوزده",
]

_FARSI_TENS = [
    "",
    "ده",
    "بیست",
    "سی",
    "چهل",
    "پنجاه",
    "شصت",
    "هفتاد",
    "هشتاد",
    "نود",
]

_FARSI_HUNDREDS = [
    "",
    "صد",
    "دویست",
    "سیصد",
    "چهارصد",
    "پانصد",
    "ششصد",
    "هفتصد",
    "هشتصد",
    "نهصد",
]

_FARSI_BIG = [
    '',
    'هزار',
    'میلیون',
    "میلیارد",
    'تریلیون',
    "تریلیارد",
]

_FORMAL_VARIANT = {
    'هفده': 'هیفده',
    'هجده': 'هیجده',
    'شانزده': 'شونزده',
    'پانزده': 'پونزده',
}

_FARSI_FRAC = ["", "ده", "صد"]
_FARSI_FRAC_BIG = ["", "هزار", "میلیونی", "میلیاردی"]
_FARSI_SEPERATOR = ' و '


class NumberVariantFA(IntEnum):
    CONVERSATIONAL = 0
    FORMAL = 1


def _is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def _parse_sentence(text):
    for key, value in _FORMAL_VARIANT.items():
        text = text.replace(key, value)
    ar = text.split()
    result = []
    current_number = 0
    current_words = []
    s = 0
    mode = 'init'

    def finish_num():
        nonlocal current_number
        nonlocal s
        nonlocal result
        nonlocal mode
        nonlocal current_words
        current_number += s
        if current_number != 0:
            result.append((current_number, current_words))
        s = 0
        current_number = 0
        current_words = []
        mode = 'init'

    for x in ar:
        if x == "و":
            if mode == 'num_ten' or mode == 'num_hundred' or mode == 'num_one':
                mode += '_va'
                current_words.append(x)
            elif mode == 'num':
                current_words.append(x)
            else:
                finish_num()
                result.append(x)
        elif x == "نیم":
            current_words.append(x)
            current_number += 0.5
            finish_num()
        elif x in _FARSI_ONES:
            t = _FARSI_ONES.index(x)
            if mode != 'init' and mode != 'num_hundred_va' and mode != 'num':
                if not (t < 10 and mode == 'num_ten_va'):
                    finish_num()
            current_words.append(x)
            s += t
            mode = 'num_one'
        elif x in _FARSI_TENS:
            if mode != 'init' and mode != 'num_hundred_va' and mode != 'num':
                finish_num()
            current_words.append(x)
            s += _FARSI_TENS.index(x) * 10
            mode = 'num_ten'
        elif x in _FARSI_HUNDREDS:
            if mode != 'init' and mode != 'num':
                finish_num()
            current_words.append(x)
            s += _FARSI_HUNDREDS.index(x) * 100
            mode = 'num_hundred'
        elif x in _FARSI_BIG:
            current_words.append(x)
            d = _FARSI_BIG.index(x)
            if mode == 'init' and d == 1:
                s = 1
            s *= 10 ** (3 * d)
            current_number += s
            s = 0
            mode = 'num'
        elif _is_number(x):
            current_words.append(x)
            current_number = float(x)
            finish_num()
        else:
            finish_num()
            result.append(x)
    if mode[:3] == 'num':
        finish_num()
    return result


def extract_numbers_fa(text, short_scale=True, ordinals=False):
    """
        Takes in a string and extracts a list of numbers.

    Args:
        text (str): the string to extract a number from
        short_scale (bool): Use "short scale" or "long scale" for large
            numbers -- over a million.  The default is short scale, which
            is now common in most English speaking countries.
            See https://en.wikipedia.org/wiki/Names_of_large_numbers
        ordinals (bool): consider ordinal numbers, e.g. third=3 instead of 1/3
    Returns:
        list: list of extracted numbers as floats
    """

    ar = _parse_sentence(text)
    result = []
    for x in ar:
        if type(x) == tuple:
            result.append(x[0])
    return result


def extract_number_fa(text, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    x = extract_numbers_fa(text, ordinals=ordinals)
    if (len(x) == 0):
        return False
    return x[0]


def nice_number_fa(number, speech=True, denominators=range(1, 21),
                   variant: NumberVariantFA = NumberVariantFA.CONVERSATIONAL):
    """ Farsi helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 and a half" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_FA[den]
    if whole == 0:
        if num == 1:
            return_string = 'یک {}'.format(den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1:
        return_string = '{} و یک {}'.format(whole, den_str)
    else:
        return_string = '{} و {} {}'.format(whole, num, den_str)
    return return_string


def _float2tuple(value, _precision):
    pre = int(value)

    post = abs(value - pre) * 10 ** _precision
    if abs(round(post) - post) < 0.01:
        # We generally floor all values beyond our precision (rather than
        # rounding), but in cases where we have something like 1.239999999,
        # which is probably due to python's handling of floats, we actually
        # want to consider it as 1.24 instead of 1.23
        post = int(round(post))
    else:
        post = int(math.floor(post))

    while post != 0:
        x, y = divmod(post, 10)
        if y != 0:
            break
        post = x
        _precision -= 1

    return pre, post, _precision


def _cardinal3(number):
    if (number < 19):
        return _FARSI_ONES[number]
    if (number < 100):
        x, y = divmod(number, 10)
        if y == 0:
            return _FARSI_TENS[x]
        return _FARSI_TENS[x] + _FARSI_SEPERATOR + _FARSI_ONES[y]
    x, y = divmod(number, 100)
    if y == 0:
        return _FARSI_HUNDREDS[x]
    return _FARSI_HUNDREDS[x] + _FARSI_SEPERATOR + _cardinal3(y)


def _cardinalPos(number):
    x = number
    res = ''
    for b in _FARSI_BIG:
        x, y = divmod(x, 1000)
        if (y == 0):
            continue
        yx = _cardinal3(y)
        if y == 1 and b == 'هزار':
            yx = b
        elif b != '':
            yx += ' ' + b
        if (res == ''):
            res = yx
        else:
            res = yx + _FARSI_SEPERATOR + res
    return res


def _fractional(number, l):
    if (number / 10 ** l == 0.5):
        return "نیم"
    x = _cardinalPos(number)
    ld3, lm3 = divmod(l, 3)
    ltext = (_FARSI_FRAC[lm3] + " " + _FARSI_FRAC_BIG[ld3]).strip() + 'م'
    return x + " " + ltext


def _to_ordinal(number):
    r = _to_cardinal(number, 0)
    if (r[-1] == 'ه' and r[-2] == 'س'):
        return r[:-1] + 'وم'
    return r + 'م'


def _to_ordinal_num(value):
    return str(value) + "م"


def _to_cardinal(number, places):
    if number < 0:
        return "منفی " + _to_cardinal(-number, places)
    if (number == 0):
        return "صفر"
    x, y, l = _float2tuple(number, places)
    if y == 0:
        return _cardinalPos(x)
    if x == 0:
        return _fractional(y, l)
    return _cardinalPos(x) + _FARSI_SEPERATOR + _fractional(y, l)


def pronounce_number_fa(number, places=2, scientific=False,
                        ordinals=False,
                        variant: NumberVariantFA = NumberVariantFA.CONVERSATIONAL):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    num = number
    # deal with infinity
    if num == float("inf"):
        return "بینهایت"
    elif num == float("-inf"):
        return "منفی بینهایت"
    if scientific:
        if number == 0:
            return "صفر"
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            return '{}{} ضرب در ده به توان {}{}'.format(
                'منفی ' if float(n) < 0 else '',
                pronounce_number_fa(
                    abs(float(n)), places, False, ordinals=False),
                'منفی ' if power < 0 else '',
                pronounce_number_fa(abs(power), places, False, ordinals=False))
    if ordinals:
        return _to_ordinal(number)
    return _to_cardinal(number, places)



from ovos_number_parser.util import (convert_to_mixed_fraction)
from math import floor
_NUM_STRING_HU = {
    0: 'nulla',
    1: 'egy',
    2: 'kettő',
    3: 'három',
    4: 'négy',
    5: 'öt',
    6: 'hat',
    7: 'hét',
    8: 'nyolc',
    9: 'kilenc',
    10: 'tíz',
    11: 'tizenegy',
    12: 'tizenkettő',
    13: 'tizenhárom',
    14: 'tizennégy',
    15: 'tizenöt',
    16: 'tizenhat',
    17: 'tizenhét',
    18: 'tizennyolc',
    19: 'tizenkilenc',
    20: 'húsz',
    30: 'harminc',
    40: 'negyven',
    50: 'ötven',
    60: 'hatvan',
    70: 'hetven',
    80: 'nyolcvan',
    90: 'kilencven',
    100: 'száz'
}

# Hungarian uses "long scale"
#    https://en.wikipedia.org/wiki/Long_and_short_scales
# Currently, numbers are limited to 1000000000000000000000000,
# but _NUM_POWERS_OF_TEN can be extended to include additional number words

_NUM_POWERS_OF_TEN = [
    '', 'ezer', 'millió', 'milliárd', 'billió', 'billiárd', 'trillió',
    'trilliárd'
]

_FRACTION_STRING_HU = {
    2: 'fél',
    3: 'harmad',
    4: 'negyed',
    5: 'ötöd',
    6: 'hatod',
    7: 'heted',
    8: 'nyolcad',
    9: 'kilenced',
    10: 'tized',
    11: 'tizenegyed',
    12: 'tizenketted',
    13: 'tizenharmad',
    14: 'tizennegyed',
    15: 'tizenötöd',
    16: 'tizenhatod',
    17: 'tizenheted',
    18: 'tizennyolcad',
    19: 'tizenkilenced',
    20: 'huszad'
}

# Numbers below 2 thousand are written in one word in Hungarian
# Numbers above 2 thousand are separated by hyphens
# In some circumstances it may better to seperate individual words
# Set _EXTRA_SPACE_HU=" " for separating numbers below 2 thousand (
# orthographically incorrect)
# Set _EXTRA_SPACE_HU="" for correct spelling, this is standard

# _EXTRA_SPACE_HU = " "
_EXTRA_SPACE_HU = ""


def _get_vocal_type_hu(word):
    # checks the vocal attributes of a word
    vowels_high = len([char for char in word if char in 'eéiíöőüű'])
    vowels_low = len([char for char in word if char in 'aáoóuú'])
    if vowels_high != 0 and vowels_low != 0:
        return 2                                   # 2: type is mixed
    return 0 if vowels_high == 0 else 1            # 0: type is low, 1: is high


def nice_number_hu(number, speech=True, denominators=range(1, 21)):
    """ Hungarian helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 és fél" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3)).replace(".", ",")

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_HU[den]
    if whole == 0:
        if num == 1:
            one = 'egy ' if den != 2 else ''
            return_string = '{}{}'.format(one, den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1:
        pointOne = 'egész egy' if den != 2 else 'és'
        return_string = '{} {} {}'.format(whole, pointOne, den_str)
    else:
        return_string = '{} egész {} {}'.format(whole, num, den_str)
    return return_string


def pronounce_number_hu(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """

    # TODO short_scale, scientific and ordinals
    # currently ignored

    def pronounce_triplet_hu(num):
        result = ""
        num = floor(num)
        if num > 99:
            hundreds = floor(num / 100)
            if hundreds > 0:
                hundredConst = _EXTRA_SPACE_HU + 'száz' + _EXTRA_SPACE_HU
                if hundreds == 1:
                    result += hundredConst
                elif hundreds == 2:
                    result += 'két' + hundredConst
                else:
                    result += _NUM_STRING_HU[hundreds] + hundredConst
                num -= hundreds * 100
        if num == 0:
            result += ''  # do nothing
        elif num <= 20:
            result += _NUM_STRING_HU[num]  # + _EXTRA_SPACE_DA
        elif num > 20:
            ones = num % 10
            tens = num - ones
            if tens > 0:
                if tens != 20:
                    result += _NUM_STRING_HU[tens] + _EXTRA_SPACE_HU
                else:
                    result += "huszon" + _EXTRA_SPACE_HU
            if ones > 0:
                result += _NUM_STRING_HU[ones] + _EXTRA_SPACE_HU
        return result

    def pronounce_whole_number_hu(num, scale_level=0):
        if num == 0:
            return ''

        num = floor(num)
        result = ''
        last_triplet = num % 1000

        if last_triplet == 1:
            if scale_level == 0:
                if result != '':
                    result += '' + "egy"
                else:
                    result += "egy"
            elif scale_level == 1:
                result += _EXTRA_SPACE_HU + \
                          _NUM_POWERS_OF_TEN[1] + _EXTRA_SPACE_HU
            else:
                result += "egy" + _NUM_POWERS_OF_TEN[scale_level]
        elif last_triplet > 1:
            result += pronounce_triplet_hu(last_triplet)
            if scale_level != 0:
                result = result.replace(_NUM_STRING_HU[2], 'két')
            if scale_level == 1:
                result += _NUM_POWERS_OF_TEN[1] + _EXTRA_SPACE_HU
            if scale_level >= 2:
                result += _NUM_POWERS_OF_TEN[scale_level]
            if scale_level > 0:
                result += '-'

        num = floor(num / 1000)
        scale_level += 1
        return pronounce_whole_number_hu(num,
                                         scale_level) + result

    result = ""
    if abs(number) >= 1000000000000000000000000:  # cannot do more than this
        return str(number)
    elif number == 0:
        return str(_NUM_STRING_HU[0])
    elif number < 0:
        return "mínusz " + pronounce_number_hu(abs(number), places)
    else:
        if number == int(number):
            return pronounce_whole_number_hu(number).strip('-')
        else:
            whole_number_part = floor(number)
            fractional_part = number - whole_number_part
            if whole_number_part == 0:
                result += _NUM_STRING_HU[0]
            result += pronounce_whole_number_hu(whole_number_part)
            if places > 0:
                result += " egész "
                fraction = pronounce_whole_number_hu(
                    round(fractional_part * 10 ** places))
                result += fraction.replace(_NUM_STRING_HU[2], 'két')
                fraction_suffixes = [
                    'tized', 'század', 'ezred', 'tízezred', 'százezred']
                if places <= len(fraction_suffixes):
                    result += ' ' + fraction_suffixes[places - 1]
            return result


def pronounce_ordinal_hu(number):
    """
    This function pronounces a number as an ordinal

    1 -> first
    2 -> second

    Args:
        number (int): the number to format
    Returns:
        (str): The pronounced number string.
    """
    ordinals = ["nulladik", "első", "második", "harmadik", "negyedik",
                "ötödik", "hatodik", "hetedik", "nyolcadik", "kilencedik",
                "tizedik"]
    big_ordinals = ["", "ezredik", "milliomodik"]

    # only for whole positive numbers including zero
    if number < 0 or number != int(number):
        return number
    elif number < 11:
        return ordinals[number]
    else:
        # concatenate parts and inflect them accordingly
        root = pronounce_number_hu(number)
        vtype = _get_vocal_type_hu(root)
        last_digit = number - floor(number / 10) * 10
        if root == "húsz":
            root = "husz"
        if number % 1000000 == 0:
            return root.replace(_NUM_POWERS_OF_TEN[2], big_ordinals[2])
        if number % 1000 == 0:
            return root.replace(_NUM_POWERS_OF_TEN[1], big_ordinals[1])
        if last_digit == 1:
            return root + "edik"
        elif root[-1] == 'ő':
            return root[:-1] + 'edik'
        elif last_digit != 0:
            return ordinals[last_digit].join(
                root.rsplit(_NUM_STRING_HU[last_digit], 1))
        return root + "edik" if vtype == 1 else root + "adik"



#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict

from ovos_number_parser.util import convert_to_mixed_fraction, is_numeric, look_for_fractions, \
    invert_dict, ReplaceableNumber, partition_list, tokenize, Token

_NUM_STRING_PL = {
    0: 'zero',
    1: 'jeden',
    2: 'dwa',
    3: 'trzy',
    4: 'cztery',
    5: 'pięć',
    6: 'sześć',
    7: 'siedem',
    8: 'osiem',
    9: 'dziewięć',
    10: 'dziesięć',
    11: 'jedenaście',
    12: 'dwanaście',
    13: 'trzynaście',
    14: 'czternaście',
    15: 'piętnaście',
    16: 'szesnaście',
    17: 'siedemnaście',
    18: 'osiemnaście',
    19: 'dziewiętnaście',
    20: 'dwadzieścia',
    30: 'trzydzieści',
    40: 'czterdzieści',
    50: 'pięćdziesiąt',
    60: 'sześćdziesiąt',
    70: 'siedemdziesiąt',
    80: 'osiemdziesiąt',
    90: 'dziewięćdziesiąt',
    100: 'sto',
    200: 'dwieście',
    300: 'trzysta',
    400: 'czterysta',
    500: 'pięćset',
    600: 'sześćset',
    700: 'siedemset',
    800: 'osiemset',
    900: 'dziewięćset',
}

_FRACTION_STRING_PL = {
    1: 'jedna',
    2: 'druga',
    3: 'trzecia',
    4: 'czwarta',
    5: 'piąta',
    6: 'szósta',
    7: 'siódma',
    8: 'ósma',
    9: 'dziewiąta',
    10: 'dziesiąta',
    11: 'jedenasta',
    12: 'dwunasta',
    13: 'trzynasta',
    14: 'czternasta',
    15: 'piętnasta',
    16: 'szesnasta',
    17: 'siedemnasta',
    18: 'osiemnasta',
    19: 'dziewiętnasta',
    20: 'dwudziesta',
    30: 'trzydziesta',
    40: 'czterdziesta',
    50: 'pięćdziesiąta',
    60: 'sześćdziesiąta',
    70: 'siedemdziesiąta',
    80: 'osiemdziesiąta',
    90: 'dziewięćdziesiąta',
    100: 'setna',
    200: 'dwusetna',
    300: 'trzysetna',
    400: 'czterysetna',
    500: 'pięćsetna',
    600: 'sześćsetna',
    700: 'siedemsetna',
    800: 'osiemsetna',
    900: 'dziewięćsetna',
    1000: 'tysięczna',
}

_SHORT_SCALE_PL = OrderedDict([
    (100, 'sto'),
    (200, 'dwieście'),
    (300, 'trzysta'),
    (400, 'czterysta'),
    (500, 'pięćset'),
    (600, 'sześćset'),
    (700, 'siedemset'),
    (800, 'osiemset'),
    (900, 'dziewięćset'),
    (1000, 'tysiąc'),
    (1000000, 'milion'),
    (1e9, "miliard"),
    (1e12, 'bilion'),
    (1e15, "biliard"),
    (1e18, "trylion"),
    (1e21, "sekstilion"),
    (1e24, "kwadrylion"),
    (1e27, "kwadryliard"),
    (1e30, "kwintylion"),
    (1e33, "kwintyliard"),
    (1e36, "sekstylion"),
    (1e39, "sekstyliard"),
    (1e42, "septylion"),
    (1e45, "septyliard"),
    (1e48, "oktylion"),
    (1e51, "oktyliard"),
    (1e54, "nonilion"),
    (1e57, "noniliard"),
    (1e60, "decylion"),
    (1e63, "decyliard"),
    (1e66, "undecylion"),
    (1e69, "undecyliard"),
    (1e72, "duodecylion"),
    (1e75, "duodecyliard"),
    (1e78, "tredecylion"),
    (1e81, "tredecyliard"),
    (1e84, "kwartyduodecylion"),
    (1e87, "kwartyduodecyliard"),
    (1e90, "kwintyduodecylion"),
    (1e93, "kwintyduodecyliard"),
    (1e96, "seksdecylion"),
    (1e99, "seksdecyliard"),
    (1e102, "septydecylion"),
    (1e105, "septydecyliard"),
    (1e108, "oktodecylion"),
    (1e111, "oktodecyliard"),
    (1e114, "nondecylion"),
    (1e117, "nondecyliard"),
    (1e120, "wigintylion"),
    (1e123, "wigintyliard"),
    (1e153, "quinquagintylion"),
    (1e183, "trycyliard"),
    (1e213, "septuagintylion"),
    (1e243, "kwadragiliard"),
    (1e273, "nonagintylion"),
    (1e303, "centezylion"),
    (1e306, "uncentylion"),
    (1e309, "duocentylion"),
    (1e312, "trescentylion"),
    (1e333, "decicentylion"),
    (1e336, "undecicentylion"),
    (1e363, "viginticentylion"),
    (1e366, "unviginticentylion"),
    (1e393, "trigintacentylion"),
    (1e423, "quadragintacentylion"),
    (1e453, "quinquagintacentylion"),
    (1e483, "sexagintacentylion"),
    (1e513, "septuagintacentylion"),
    (1e543, "ctogintacentylion"),
    (1e573, "nonagintacentylion"),
    (1e603, "centyliard"),
    (1e903, "trecentylion"),
    (1e1203, "quadringentylion"),
    (1e1503, "quingentylion"),
    (1e1803, "sescentylion"),
    (1e2103, "septingentylion"),
    (1e2403, "octingentylion"),
    (1e2703, "nongentylion"),
    (1e3003, "milinylion")
])

_ORDINAL_BASE_PL = {
    1: 'pierwszy',
    2: 'drugi',
    3: 'trzeci',
    4: 'czwarty',
    5: 'piąty',
    6: 'szósty',
    7: 'siódmy',
    8: 'ósmy',
    9: 'dziewiąty',
    10: 'dziesiąty',
    11: 'jedenasty',
    12: 'dwunasty',
    13: 'trzynasty',
    14: 'czternasty',
    15: 'piętnasty',
    16: 'szesnasty',
    17: 'siedemnasty',
    18: 'osiemnasty',
    19: 'dziewiętnasty',
    20: 'dwudziesty',
    30: 'trzydziesty',
    40: "czterdziesty",
    50: "pięćdziesiąty",
    60: "sześćdziesiąty",
    70: "siedemdziesiąty",
    80: "osiemdziesiąty",
    90: "dziewięćdziesiąty",
    1e2: "setny",
    1e3: "tysięczny"
}

_SHORT_ORDINAL_PL = {
    1e6: "milionowy",
    1e9: "miliardowy",
    1e12: "bilionowy",
    1e15: "biliardowy",
    1e18: "trylionowy",
    1e21: "tryliardowy",
    1e24: "kwadrylionowy",
    1e27: "kwadryliardowy",
    1e30: "kwintylionowy",
    1e33: "kwintyliardowy",
    1e36: "sektylionowy",
    1e42: "septylionowy",
    1e48: "oktylionowy",
    1e54: "nonylionowy",
    1e60: "decylionowy"
    # TODO > 1e-33
}
_SHORT_ORDINAL_PL.update(_ORDINAL_BASE_PL)

_ALT_ORDINALS_PL = {
    1: 'pierwszej',
    2: 'drugiej',
    3: 'trzeciej',
    4: 'czwartej',
    5: 'piątej',
    6: 'szóstej',
    7: 'siódmej',
    8: 'ósmej',
    9: 'dziewiątej',
    10: 'dziesięcio',
    11: 'jedenasto',
    12: 'dwunasto',
    13: 'trzynasto',
    14: 'czternasto',
    15: 'piętnasto',
    16: 'szesnasto',
    17: 'siedemnasto',
    18: 'osiemnasto',
    19: 'dziewiętnasto',
    20: 'dwudziesto',
    30: 'trzydziesto',
    40: 'czterdziesto',
    50: 'pięćdziesiecio',
    60: 'sześćdziesięcio',
    70: 'siedemdziesięcio',
    80: 'osiemdziesięcio',
    90: 'dziewięćdziesięcio',
}


def nice_number_pl(number, speech=True, denominators=range(1, 21)):
    """ English helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 and a half" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_PL[den]
    if whole == 0:
        return_string = '{} {}'.format(num, den_str)
    else:
        return_string = '{} i {} {}'.format(whole, num, den_str)
    if num > 1:
        return_string = return_string[:-1] + 'e'
    return return_string


def pronounce_number_pl(num, places=2, short_scale=True, scientific=False,
                        ordinals=False, scientific_run=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    # deal with infinity
    if num == float("inf"):
        return "nieskończoność"
    elif num == float("-inf"):
        return "minus nieskończoność"
    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            if ordinals:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} razy dziesięć do {}{} potęgi'.format(
                    'minus ' if float(n) < 0 else '',
                    pronounce_number_pl(
                        abs(float(n)), places, short_scale, False, ordinals=False, scientific_run=True),
                    'minus ' if power < 0 else '',
                    pronounce_number_pl(abs(power), places, short_scale, False, ordinals=True, scientific_run=True))
            else:
                # This handles negatives of powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} razy dziesięć do potęgi {}{}'.format(
                    'minus ' if float(n) < 0 else '',
                    pronounce_number_pl(
                        abs(float(n)), places, short_scale, False),
                    'minus ' if power < 0 else '',
                    pronounce_number_pl(abs(power), places, short_scale, False))

    number_names = _NUM_STRING_PL.copy()
    number_names.update(_SHORT_SCALE_PL)

    digits = [number_names[n] for n in range(0, 20)]
    if ordinals:
        tens = [_SHORT_ORDINAL_PL[n] for n in range(10, 100, 10)]
    else:
        tens = [number_names[n] for n in range(10, 100, 10)]
    hundreds = [_SHORT_SCALE_PL[n] for n in _SHORT_SCALE_PL.keys()]

    # deal with negatives
    result = ""
    if num < 0:
        result = "minus "
    num = abs(num)

    # check for a direct match
    if num in number_names and not ordinals:
        result += number_names[num]
    else:
        def _sub_thousand(n, ordinals=False, iteration=0):
            assert 0 <= n <= 999

            _, n_mod = divmod(n, 10)
            if iteration > 0 and n in _ALT_ORDINALS_PL and ordinals:
                return _ALT_ORDINALS_PL[n]
            elif n in _SHORT_ORDINAL_PL and ordinals:
                return _SHORT_ORDINAL_PL[n] if not scientific_run \
                    else _ALT_ORDINALS_PL[n]
            if n <= 19:
                return digits[n] if not scientific_run or not ordinals \
                    else digits[n][:-1] + "ej"
            elif n <= 99:
                q, r = divmod(n, 10)
                tens_text = tens[q - 1]
                if scientific_run:
                    tens_text = tens_text[:-1] + "ej"
                return tens_text + (" " + _sub_thousand(r, ordinals) if r
                                    else "")
            else:
                q, r = divmod(n, 100)
                digit_name = digits[q]
                if q * 100 in _NUM_STRING_PL:
                    digit_name = _NUM_STRING_PL[q * 100]

                return digit_name + (
                    " " + _sub_thousand(r, ordinals) if r else "")

        def _short_scale(n):
            if n >= max(_SHORT_SCALE_PL.keys()):
                return "nieskończoność"
            ordi = ordinals

            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000)):
                if not z:
                    continue
                number = _sub_thousand(z, ordi, iteration=i)

                if i:
                    if i >= len(hundreds):
                        return ""
                    number += " "
                    if ordi:
                        if i * 1000 in _SHORT_ORDINAL_PL:
                            if z == 1:
                                number = _SHORT_ORDINAL_PL[i * 1000]
                            else:
                                number += _SHORT_ORDINAL_PL[i * 1000]
                        else:
                            if n not in _SHORT_SCALE_PL:
                                num = int("1" + "0" * (len(str(n)) - 2))

                                number += _SHORT_SCALE_PL[num] + "owa"
                            else:
                                number = _SHORT_SCALE_PL[n] + "ty"
                    else:
                        hundreds_text = _SHORT_SCALE_PL[float(pow(1000, i))]
                        if z != 1:
                            _, z_mod = divmod(z, 10)
                            _, z_mod_tens = divmod(z, 100)
                            n_main, _ = divmod(z_mod_tens, 10)
                            if i == 1:
                                if n_main != 1 and 5 > z_mod > 0:
                                    hundreds_text += "e"
                                else:
                                    hundreds_text = "tysięcy"
                            elif i > 1:
                                hundreds_text += "y" if 5 > z_mod > 0 else "ów"

                        number += hundreds_text
                res.append(number)
                ordi = False

            return ", ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        result += _short_scale(num)

    # deal with scientific notation unpronounceable as number
    if not result and "e" in str(num):
        return pronounce_number_pl(num, places, short_scale, scientific=True)
    # Deal with fractional part
    elif not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "minus " or not result):
            result += "zero"
        result += " przecinek"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result


def generate_plurals_pl(originals):
    """
    Return a new set or dict containing the plural form of the original values,

    In English this means all with 's' appended to them.

    Args:
        originals set(str) or dict(str, any): values to pluralize

    Returns:
        set(str) or dict(str, any)

    """
    if isinstance(originals, dict):
        result = {key + 'y': value for key, value in originals.items()}
        result = {**result, **{key + 'ów': value for key, value in originals.items()}}
        result = {**result, **{'tysiące': 1000, 'tysięcy': 1000}}

        return result

    result = {value + "y" for value in originals}
    result = result.union({value + "ów" for value in originals})
    result = result.union({'tysiące', 'tysięcy'})

    return result


def generate_fractions_pl(fractions):
    '''Returns a list of all fraction combinations. E.g.:
    trzecia, trzecich, trzecie
    czwarta, czwarte, czwartych

    :param fractions: Existing fractions
    :return: Fractions with add suffixes
    '''

    result = {**fractions}
    for k, v in fractions.items():
        k_no_last = k[:-1]
        result[k_no_last + 'e'] = v
        if k_no_last[-1:] == 'i':
            result[k_no_last + 'ch'] = v
        else:
            result[k_no_last + 'ych'] = v

    for k, v in _SHORT_ORDINAL_PL.items():
        result[v[:-1] + 'a'] = k

    result['jedno'] = 1
    result['czwartego'] = 4

    return result


# negate next number (-2 = 0 - 2)
_NEGATIVES = {"ujemne", "minus"}

# sum the next number (twenty two = 20 + 2)
_SUMS = {'dwadzieścia', '20', 'trzydzieści', '30', 'czterdzieści', '40', 'pięćdziesiąt', '50',
         'sześćdziesiąt', '60', 'siedemdziesiąt', '70', 'osiemdziesiąt', '80', 'dziewięćdziesiąt', '90'}

_MULTIPLIES_SHORT_SCALE_PL = generate_plurals_pl(_SHORT_SCALE_PL.values())

# split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
_FRACTION_MARKER = {'i'}

# decimal marker ( 1 point 5 = 1 + 0.5)
_DECIMAL_MARKER = {'kropka', 'przecinek'}

_STRING_NUM_PL = invert_dict(_NUM_STRING_PL)
_STRING_NUM_PL.update(generate_plurals_pl(_STRING_NUM_PL))
_STRING_NUM_PL.update({
    'pół': 0.5,
    'połówka': 0.5,
    'połowa': 0.5,
})

_STRING_SHORT_ORDINAL_PL = invert_dict(_SHORT_ORDINAL_PL)

_REV_FRACTITONS = generate_fractions_pl(invert_dict(_FRACTION_STRING_PL))


def numbers_to_digits_pl(text, short_scale=True, ordinals=False):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numbers should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)

    Returns:
        str
        The original text, with numbers subbed in where appropriate.

    """
    text = text.lower()
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_pl(tokens, short_scale, ordinals)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_pl(tokens, short_scale=True,
                                  ordinals=False, fractional_numbers=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_pl(tokens, short_scale,
                                         ordinals, fractional_numbers)

        if not to_replace:
            break

        results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_pl(tokens, short_scale=True,
                                 ordinals=False, fractional_numbers=True):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = \
        _extract_number_with_text_pl_helper(tokens, short_scale,
                                            ordinals, fractional_numbers)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_pl_helper(tokens,
                                        short_scale=True, ordinals=False,
                                        fractional_numbers=True):
    """
    Helper for _extract_number_with_text_en.

    This contains the real logic for parsing, but produces
    a result that needs a little cleaning (specific, it may
    contain leading articles that can be trimmed off).

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:

    Returns:
        int or float, [Tokens]

    """
    if fractional_numbers:
        fraction, fraction_text = \
            _extract_fraction_with_text_pl(tokens, short_scale, ordinals)
        if fraction:
            return fraction, fraction_text

        decimal, decimal_text = \
            _extract_decimal_with_text_pl(tokens, short_scale, ordinals)
        if decimal:
            return decimal, decimal_text

    return _extract_whole_number_with_text_pl(tokens, short_scale, ordinals)


def _extract_fraction_with_text_pl(tokens, short_scale, ordinals):
    """
    Extract fraction numbers from a string.

    This function handles text such as '2 and 3/4'. Note that "one half" or
    similar will be parsed by the whole number function.

    Args:
        tokens [Token]: words and their indexes in the original string.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (int or float, [Token])
        The value found, and the list of relevant tokens.
        (None, None) if no fraction value is found.

    """
    for c in _FRACTION_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_pl(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_pl(partitions[2], short_scale,
                                              ordinals, fractional_numbers=True)

            if not numbers1 or not numbers2:
                return None, None

            # ensure first is not a fraction and second is a fraction
            num1 = numbers1[-1]
            num2 = numbers2[0]
            if num1.value >= 1 and 0 < num2.value < 1:
                return num1.value + num2.value, \
                       num1.tokens + partitions[1] + num2.tokens

    return None, None


def _extract_decimal_with_text_pl(tokens, short_scale, ordinals):
    """
    Extract decimal numbers from a string.

    This function handles text such as '2 point 5'.

    Notes:
        While this is a helper for extractnumber_en, it also depends on
        extractnumber_en, to parse out the components of the decimal.

        This does not currently handle things like:
            number dot number number number

    Args:
        tokens [Token]: The text to parse.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (float, [Token])
        The value found and relevant tokens.
        (None, None) if no decimal value is found.

    """
    for c in _DECIMAL_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_pl(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_pl(partitions[2], short_scale,
                                              ordinals, fractional_numbers=False)

            if not numbers1 or not numbers2:
                return None, None

            number = numbers1[-1]
            decimal = numbers2[0]

            # TODO handle number dot number number number
            if "." not in str(decimal.text):
                return number.value + float('0.' + str(decimal.value)), \
                       number.tokens + partitions[1] + decimal.tokens
    return None, None


def _extract_whole_number_with_text_pl(tokens, short_scale, ordinals):
    """
    Handle numbers not handled by the decimal or fraction functions. This is
    generally whole numbers. Note that phrases such as "one half" will be
    handled by this function, while "one and a half" are handled by the
    fraction function.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:

    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.

    """
    multiplies, string_num_ordinal, string_num_scale = \
        _initialize_number_data(short_scale)

    number_words = []  # type: [Token]
    val = False
    prev_val = None
    next_val = None
    to_sum = []
    for idx, token in enumerate(tokens):
        current_val = None
        if next_val:
            next_val = None
            continue

        word = token.word

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""

        if is_numeric(word[:-1]) and word.endswith('.'):
            # explicit ordinals, 1., 2., 3., 4.... N.
            word = word[:-1]

        word = normalize_word_pl(word)

        if word not in string_num_scale and \
                word not in _STRING_NUM_PL and \
                word not in _SUMS and \
                word not in multiplies and \
                not (ordinals and word in string_num_ordinal) and \
                not is_numeric(word) and \
                not is_fractional_pl(word) and \
                not look_for_fractions(word.split('/')):
            words_only = [token.word for token in number_words]
            if number_words and not all([w in _NEGATIVES for w in words_only]):
                break
            else:
                number_words = []
                continue
        elif word not in multiplies \
                and prev_word not in multiplies \
                and prev_word not in _SHORT_SCALE_PL.values() \
                and prev_word not in _SUMS \
                and not (ordinals and prev_word in string_num_ordinal) \
                and prev_word not in _NEGATIVES:
            number_words = [token]
        elif prev_word in _SUMS and word in _SUMS:
            number_words = [token]
        else:
            number_words.append(token)

        # is this word already a number ?
        if is_numeric(word):
            if word.isdigit():  # doesn't work with decimals
                val = int(word)
            else:
                val = float(word)
            current_val = val

        # is this word the name of a number ?
        if word in _STRING_NUM_PL:
            val = _STRING_NUM_PL.get(word)
            current_val = val
        elif word in string_num_scale:
            val = string_num_scale.get(word)
            current_val = val
        elif ordinals and word in string_num_ordinal:
            val = string_num_ordinal[word]
            current_val = val

        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val
            prev_val = None

        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if prev_val:
            if (prev_word in string_num_ordinal and val and val < prev_val) or \
                    (prev_word in _STRING_NUM_PL and val and val < prev_val and val // 10 != prev_val // 10) or \
                    all([prev_word in multiplies, val < prev_val if prev_val else False]):
                val += prev_val

        if next_word in multiplies:
            prev_val = val
            continue

        # is this a spoken fraction?
        # half cup
        if val is False:
            val = is_fractional_pl(word)
            current_val = val

        # 2 fifths
        if not ordinals:
            next_val = is_fractional_pl(next_word)
            if next_val:
                if not val:
                    val = 1
                val *= next_val
                number_words.append(tokens[idx + 1])

        # is this a negative number?
        if val and prev_word and prev_word in _NEGATIVES:
            val = 0 - val

        if next_word in _STRING_NUM_PL:
            prev_val = val

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])
                number_words.append(tokens[idx + 1])
        else:
            if all([
                prev_word in _SUMS,
                word not in _SUMS,
                word not in multiplies,
                current_val >= 10]):
                # Backtrack - we've got numbers we can't sum.
                number_words.pop()
                val = prev_val
                break
            prev_val = val

            if word in multiplies and next_word not in multiplies:
                # handle long numbers
                # six hundred sixty six
                # two million five hundred thousand
                #
                # This logic is somewhat complex, and warrants
                # extensive documentation for the next coder's sake.
                #
                # The current word is a power of ten. `current_val` is
                # its integer value. `val` is our working sum
                # (above, when `current_val` is 1 million, `val` is
                # 2 million.)
                #
                # We have a dict `string_num_scale` containing [value, word]
                # pairs for "all" powers of ten: string_num_scale[10] == "ten.
                #
                # We need go over the rest of the tokens, looking for other
                # powers of ten. If we find one, we compare it with the current
                # value, to see if it's smaller than the current power of ten.
                #
                # Numbers which are not powers of ten will be passed over.
                #
                # If all the remaining powers of ten are smaller than our
                # current value, we can set the current value aside for later,
                # and begin extracting another portion of our final result.
                # For example, suppose we have the following string.
                # The current word is "million".`val` is 9000000.
                # `current_val` is 1000000.
                #
                #    "nine **million** nine *hundred* seven **thousand**
                #     six *hundred* fifty seven"
                #
                # Iterating over the rest of the string, the current
                # value is larger than all remaining powers of ten.
                #
                # The if statement passes, and nine million (9000000)
                # is appended to `to_sum`.
                #
                # The main variables are reset, and the main loop begins
                # assembling another number, which will also be appended
                # under the same conditions.
                #
                # By the end of the main loop, to_sum will be a list of each
                # "place" from 100 up: [9000000, 907000, 600]
                #
                # The final three digits will be added to the sum of that list
                # at the end of the main loop, to produce the extracted number:
                #
                #    sum([9000000, 907000, 600]) + 57
                # == 9,000,000 + 907,000 + 600 + 57
                # == 9,907,657
                #
                # >>> foo = "nine million nine hundred seven thousand six
                #            hundred fifty seven"
                # >>> extract_number(foo)
                # 9907657

                time_to_sum = True
                for other_token in tokens[idx + 1:]:
                    if other_token.word in multiplies:
                        if string_num_scale[other_token.word] >= current_val:
                            time_to_sum = False
                        else:
                            continue
                    if not time_to_sum:
                        break
                if time_to_sum:
                    to_sum.append(val)
                    val = 0
                    prev_val = 0

    if val is not None and to_sum:
        val += sum(to_sum)

    return val, number_words


def _initialize_number_data(short_scale):
    """
    Generate dictionaries of words to numbers, based on scale.

    This is a helper function for _extract_whole_number.

    Args:
        short_scale boolean:

    Returns:
        (set(str), dict(str, number), dict(str, number))
        multiplies, string_num_ordinal, string_num_scale

    """
    multiplies = _MULTIPLIES_SHORT_SCALE_PL

    string_num_scale = invert_dict(_SHORT_SCALE_PL)
    string_num_scale.update(generate_plurals_pl(string_num_scale))
    return multiplies, _STRING_SHORT_ORDINAL_PL, string_num_scale


def extract_number_pl(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    return _extract_number_with_text_pl(tokenize(text.lower()),
                                        True, ordinals).value


def is_fractional_pl(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    lower_input = input_str.lower()
    if lower_input in _REV_FRACTITONS:
        return 1.0 / _REV_FRACTITONS[lower_input]

    return False


def normalize_word_pl(word):
    if word.startswith('jedn'):
        suffix = 'ą', 'ej', 'ym'
        if word.endswith(suffix):
            return 'jedna'
    if word == 'dwie':
        return 'dwa'

    return word



from ovos_number_parser.util import (convert_to_mixed_fraction, look_for_fractions,
                                     is_numeric)

_NUM_STRING_EU = {
    "zero": 0,
    "bat": 1,
    "bi": 2,
    "hiru": 3,
    "lau": 4,
    "bost": 5,
    "sei": 6,
    "zazpi": 7,
    "zortzi": 8,
    "bederatzi": 9,
    "hamar": 10,
    "hamaika": 11,
    "hamabi": 12,
    "hamahiru": 13,
    "hamalau": 14,
    "hamabost": 15,
    "hamasei": 16,
    "hamazazpi": 17,
    "hemezortzi": 18,
    "hemeretzi": 19,
    "hogei": 20,
    "hogeita hamar": 30,
    "hogeita hamaika": 31,
    "berrogei": 40,
    "berrogeita hamar": 50,
    "hirurogei": 60,
    "hirurogeita hamar": 70,
    "laurogei": 80,
    "laurogeita hamar": 90,
    "ehun": 100,
    "berrehun": 200,
    "hirurehun": 300,
    "laurehun": 400,
    "bostehun": 500,
    "seirehun": 600,
    "zazpirehun": 700,
    "zortzirehun": 800,
    "bederatzirehun": 900,
    "mila": 1000}

NUM_STRING_EU = {
    0: 'zero',
    1: 'bat',
    2: 'bi',
    3: 'hiru',
    4: 'lau',
    5: 'bost',
    6: 'sei',
    7: 'zazpi',
    8: 'zortzi',
    9: 'bederatzi',
    10: 'hamar',
    11: 'hamaika',
    12: 'hamabi',
    13: 'hamahiru',
    14: 'hamalau',
    15: 'hamabost',
    16: 'hamasei',
    17: 'hamazazpi',
    18: 'hemezortzi',
    19: 'hemeretzi',
    20: 'hogei',
    30: 'hogeita hamar',
    40: 'berrogei',
    50: 'berrogeita hamar',
    60: 'hirurogei',
    70: 'hirurogehita hamar',
    80: 'laurogei',
    90: 'laurogeita hamar',
    100: 'ehun',
    200: 'berrehun',
    300: 'hirurehun',
    400: 'laurehun',
    500: 'bostehun',
    600: 'seirehun',
    700: 'zazpirehun',
    800: 'zortzirehun',
    900: 'bederatzirehun',
    1000: 'mila'
}

FRACTION_STRING_EU = {
    2: 'erdi',
    3: 'heren',
    4: 'laurden',
    5: 'bosten',
    6: 'seiren',
    7: 'zazpiren',
    8: 'zortziren',
    9: 'bederatziren',
    10: 'hamarren',
    11: 'hamaikaren',
    12: 'hamabiren',
    13: 'hamahiruren',
    14: 'hamalauren',
    15: 'hamabosten',
    16: 'hamaseiren',
    17: 'hamazazpiren',
    18: 'hemezortziren',
    19: 'hemeretziren',
    20: 'hogeiren'
}


def nice_number_eu(number, speech=True, denominators=range(1, 21)):
    """ Euskara helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 eta erdi" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    strNumber = ""
    whole = 0
    num = 0
    den = 0

    result = convert_to_mixed_fraction(number, denominators)

    if not result:
        # Give up, just represent as a 3 decimal number
        whole = round(number, 3)
    else:
        whole, num, den = result

    if not speech:
        if num == 0:
            strNumber = '{:,}'.format(whole)
            strNumber = strNumber.replace(",", " ")
            strNumber = strNumber.replace(".", ",")
            return strNumber
        else:
            return '{} {}/{}'.format(whole, num, den)
    else:
        if num == 0:
            # if the number is not a fraction, nothing to do
            strNumber = str(whole)
            strNumber = strNumber.replace(".", ",")
            return strNumber
        den_str = FRACTION_STRING_EU[den]
        # if it is not an integer
        if whole == 0:
            # if there is no whole number
            if num == 1:
                # if numerator is 1, return "un medio", for example
                strNumber = '{} bat'.format(den_str)
            else:
                # else return "cuatro tercios", for example
                strNumber = '{} {}'.format(num, den_str)
        elif num == 1:
            # if there is a whole number and numerator is 1
            if den == 2:
                # if denominator is 2, return "1 y medio", for example
                strNumber = '{} eta {}'.format(whole, den_str)
            else:
                # else return "1 y 1 tercio", for example
                strNumber = '{} eta {} bat'.format(whole, den_str)
        else:
            # else return "2 y 3 cuarto", for example
            strNumber = '{} eta {} {}'.format(whole, num, den_str)

    return strNumber


def pronounce_number_eu(num, places=2):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'bost koma bi'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(num) >= 10000:
        # TODO: Soporta a números por encima de 1000
        return str(num)

    result = ""
    if num < 0:
        result = "minus "
    num = abs(num)

    thousands = int(num - int(num) % 1000)
    _num = num - thousands
    hundreds = int(_num - int(_num) % 100)
    _num = _num - hundreds
    tens = int(_num - _num % 10)
    ones = int(_num - tens)

    if thousands > 0:
        if thousands > 1000:
            result += NUM_STRING_EU[int(thousands / 1000)] + ' '
        result += NUM_STRING_EU[1000]
        if hundreds > 0 and tens == 0 and ones == 0:
            result += ' eta '
        elif hundreds > 0 or tens > 0 or ones > 0:
            result += ' '
    if hundreds > 0:
        result += NUM_STRING_EU[hundreds]
        if tens > 0 or ones > 0:
            result += ' eta '
    if tens or ones:
        if tens == 0 or tens == 10 or ones == 0:
            result += NUM_STRING_EU[int(_num)]
        else:
            if (tens % 20) == 10:
                ones = ones + 10
            result += NUM_STRING_EU[int(tens)].split(' ')[0].replace("ta", "") + str("ta ") + NUM_STRING_EU[int(ones)]
    if abs(num) < 1.0:
        result += NUM_STRING_EU[0]
    # Deal with decimal part, in basque is commonly used the comma
    # instead the dot. Decimal part can be written both with comma
    # and dot, but when pronounced, its pronounced "koma"
    if not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "minus " or not result):
            result += NUM_STRING_EU[0]
        result += " koma"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + NUM_STRING_EU[int(char)]

    return result


def is_fractional_eu(input_str):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        text (str): the string to check if fractional
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "fifths"

    aFrac = {"erdia": 2, "erdi": 2, "heren": 3, "laurden": 4,
             "laurdena": 4, "bosten": 5, "bostena": 5, "seiren": 6, "seirena": 6,
             "zazpiren": 7, "zapirena": 7, "zortziren": 8, "zortzirena": 8,
             "bederatziren": 9, "bederatzirena": 9, "hamarren": 10, "hamarrena": 10,
             "hamaikaren": 11, "hamaikarena": 11, "hamabiren": 12, "hamabirena": 12}

    if input_str.lower() in aFrac:
        return 1.0 / aFrac[input_str]
    if (input_str == "hogeiren" or input_str == "hogeirena"):
        return 1.0 / 20
    if (input_str == "hogeita hamarren" or input_str == "hogeita hamarrena"):
        return 1.0 / 30
    if (input_str == "ehunen" or input_str == "ehunena"):
        return 1.0 / 100
    if (input_str == "milaren" or input_str == "milarena"):
        return 1.0 / 1000
    return False


# TODO: short_scale and ordinals don't do anything here.
# The parameters are present in the function signature for API compatibility
# reasons.
#
# Returns incorrect output on certain fractional phrases like, "cuarto de dos"
def extract_number_eu(text, short_scale=True, ordinals=False):
    """
    This function prepares the given text for parsing by making
    numbers consistent, getting rid of contractions, etc.
    Args:
        text (str): the string to normalize
    Returns:
        (int) or (float): The value of extracted number

    """
    aWords = text.lower().split()
    count = 0
    result = None
    while count < len(aWords):
        val = 0
        word = aWords[count]
        next_next_word = None
        if count + 1 < len(aWords):
            next_word = aWords[count + 1]
            if count + 2 < len(aWords):
                next_next_word = aWords[count + 2]
        else:
            next_word = None

        # is current word a number?
        if word in _NUM_STRING_EU:
            val = _NUM_STRING_EU[word]
        elif word.isdigit():  # doesn't work with decimals
            val = int(word)
        elif is_numeric(word):
            val = float(word)
        elif is_fractional_eu(word):
            if next_word in _NUM_STRING_EU:
                # erdi bat, heren bat, etab
                result = _NUM_STRING_EU[next_word]
                # hurrengo hitza (bat, bi, ...) salto egin
                next_word = None
                count += 2
            elif not result:
                result = 1
                count += 1
            result = result * is_fractional_eu(word)
            continue

        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            # if (len(aPieces) == 2 and is_numeric(aPieces[0])
            #   and is_numeric(aPieces[1])):
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        if val:
            if result is None:
                result = 0
            # handle fractions
            if next_word == "en" or next_word == "ren":
                result = float(result) / float(val)
            else:
                result = val

        if next_word is None:
            break

        # number word and fraction
        ands = ["eta"]
        if next_word in ands:
            zeros = 0
            if result is None:
                count += 1
                continue
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "

            afterAndVal = extract_number_eu(newText[:-1])
            if afterAndVal:
                if result < afterAndVal or result < 20:
                    while afterAndVal > 1:
                        afterAndVal = afterAndVal / 10.0
                    for word in newWords:
                        if word == "zero" or word == "0":
                            zeros += 1
                        else:
                            break
                for _ in range(0, zeros):
                    afterAndVal = afterAndVal / 10.0
                result += afterAndVal
                break
        elif next_next_word is not None:
            if next_next_word in ands:
                newWords = aWords[count + 3:]
                newText = ""
                for word in newWords:
                    newText += word + " "
                afterAndVal = extract_number_eu(newText[:-1])
                if afterAndVal:
                    if result is None:
                        result = 0
                    result += afterAndVal
                    break

        decimals = ["puntu", "koma", ".", ","]
        if next_word in decimals:
            zeros = 0
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "
            for word in newWords:
                if word == "zero" or word == "0":
                    zeros += 1
                else:
                    break
            afterDotVal = str(extract_number_eu(newText[:-1]))
            afterDotVal = zeros * "0" + afterDotVal
            result = float(str(result) + "." + afterDotVal)
            break
        count += 1

    # Return the $str with the number related words removed
    # (now empty strings, so strlen == 0)
    # aWords = [word for word in aWords if len(word) > 0]
    # text = ' '.join(aWords)
    if "." in str(result):
        integer, dec = str(result).split(".")
        # cast float to int
        if dec == "0":
            result = int(integer)

    return result or False


# TODO Not parsing 'cero'
def eu_number_parse(words, i):
    def eu_cte(i, s):
        if i < len(words) and s == words[i]:
            return s, i + 1
        return None

    def eu_number_word(i, mi, ma):
        if i < len(words):
            v = _NUM_STRING_EU.get(words[i])
            if v and v >= mi and v <= ma:
                return v, i + 1
        return None

    def eu_number_1_99(i):
        if i >= len(words):
            return None
        r1 = eu_number_word(i, 1, 29)
        if r1:
            return r1

        composed = False
        if words[i] != "eta" and words[i][-2:] == "ta":
            composed = True
            words[i] = words[i][:-2]

        r1 = eu_number_word(i, 20, 90)

        if r1:
            v1, i1 = r1

            if composed:
                # i2 = r2[1]
                r3 = eu_number_word(i1, 1, 19)
                if r3:
                    v3, i3 = r3
                    return v1 + v3, i3
            return r1
        return None

    def eu_number_1_999(i):
        r1 = eu_number_word(i, 100, 900)
        if r1:
            v1, i1 = r1
            r2 = eu_cte(i1, "eta")
            if r2:
                i2 = r2[1]
                r3 = eu_number_1_99(i2)
                if r3:
                    v3, i3 = r3
                    return v1 + v3, i3
            else:
                return r1

        # [1-99]
        r1 = eu_number_1_99(i)
        if r1:
            return r1

        return None

    def eu_number(i):
        # check for cero
        r1 = eu_number_word(i, 0, 0)
        if r1:
            return r1

        # check for [1-999] (mil [0-999])?
        r1 = eu_number_1_999(i)
        if r1:
            v1, i1 = r1
            r2 = eu_cte(i1, "mila")
            if r2:
                i2 = r2[1]
                r3 = eu_number_1_999(i2)
                if r3:
                    v3, i3 = r3
                    return v1 * 1000 + v3, i3
                else:
                    return v1 * 1000, i2
            else:
                return r1
        return None

    return eu_number(i)



from collections import OrderedDict
from math import floor

from ovos_number_parser.util import (invert_dict, convert_to_mixed_fraction, tokenize, ReplaceableNumber, Token,
                                     look_for_fractions)

_ARTICLES = {'der', 'das', 'die', 'dem', 'den'}

# _SPOKEN_NUMBER
_NUM_STRING = {
    0: 'null',
    1: 'eins',
    2: 'zwei',
    3: 'drei',
    4: 'vier',
    5: 'fünf',
    6: 'sechs',
    7: 'sieben',
    8: 'acht',
    9: 'neun',
    10: 'zehn',
    11: 'elf',
    12: 'zwölf',
    13: 'dreizehn',
    14: 'vierzehn',
    15: 'fünfzehn',
    16: 'sechzehn',
    17: 'siebzehn',
    18: 'achtzehn',
    19: 'neunzehn',
    20: 'zwanzig',
    30: 'dreißig',
    40: 'vierzig',
    50: 'fünfzig',
    60: 'sechzig',
    70: 'siebzig',
    80: 'achtzig',
    90: 'neunzig',
    100: 'hundert',
    200: 'zweihundert',
    300: 'dreihundert',
    400: 'vierhundert',
    500: 'fünfhundert',
    600: 'sechshundert',
    700: 'siebenhundert',
    800: 'achthundert',
    900: 'neunhundert',
    1000: 'tausend',
    1000000: 'million'
}

_STRING_NUM = invert_dict(_NUM_STRING)
_STRING_NUM.update({
    'ein': 1,
    'eine': 1,
    'einer': 1,
    'eines': 1,
    'einem': 1,
    'einen': 1
})

_MONTHS = ['januar', 'februar', 'märz', 'april', 'mai', 'juni',
           'juli', 'august', 'september', 'oktober', 'november',
           'dezember']

# German uses "long scale" https://en.wikipedia.org/wiki/Long_and_short_scales
# Currently, numbers are limited to 1000000000000000000000000,
# but _NUM_POWERS_OF_TEN can be extended to include additional number words


_NUM_POWERS_OF_TEN = [
    '', 'tausend', 'Million', 'Milliarde', 'Billion', 'Billiarde', 'Trillion',
    'Trilliarde'
]

_FRACTION_STRING = {
    2: 'halb',
    3: 'drittel',
    4: 'viertel',
    5: 'fünftel',
    6: 'sechstel',
    7: 'siebtel',
    8: 'achtel',
    9: 'neuntel',
    10: 'zehntel',
    11: 'elftel',
    12: 'zwölftel',
    13: 'dreizehntel',
    14: 'vierzehntel',
    15: 'fünfzehntel',
    16: 'sechzehntel',
    17: 'siebzehntel',
    18: 'achtzehntel',
    19: 'neunzehntel',
    20: 'zwanzigstel'
}

_STRING_FRACTION = invert_dict(_FRACTION_STRING)
_STRING_FRACTION.update({
    'halb': 2,
    'halbe': 2,
    'halben': 2,
    'halbes': 2,
    'halber': 2,
    'halbem': 2
})

# Numbers below 1 million are written in one word in German, yielding very
# long words
# In some circumstances it may better to seperate individual words
# Set _EXTRA_SPACE_DA=" " for separating numbers below 1 million (
# orthographically incorrect)
# Set _EXTRA_SPACE_DA="" for correct spelling, this is standard

# _EXTRA_SPACE_DA = " "
_EXTRA_SPACE = ""

_ORDINAL_BASE = {
    "1.": "erst",
    "2.": "zweit",
    "3.": "dritt",
    "4.": "viert",
    "5.": "fünft",
    "6.": "sechst",
    "7.": "siebt",
    "8.": "acht",
    "9.": "neunt",
    "10.": "zehnt",
    "11.": "elft",
    "12.": "zwölft",
    "13.": "dreizehnt",
    "14.": "vierzehnt",
    "15.": "fünfzehnt",
    "16.": "sechzehnt",
    "17.": "siebzehnt",
    "18.": "achtzehnt",
    "19.": "neunzehnt",
    "20.": "zwanzigst",
    "21.": "einundzwanzigst",
    "22.": "zweiundzwanzigst",
    "23.": "dreiundzwanzigst",
    "24.": "vierundzwanzigst",
    "25.": "fünfundzwanzigst",
    "26.": "sechsundzwanzigst",
    "27.": "siebenundzwanzigst",
    "28.": "achtundzwanzigst",
    "29.": "neunundzwanzigst",
    "30.": "dreißigst",
    "31.": "einunddreißigst",
    "32.": "zweiunddreißigst",
    "33.": "dreiunddreißigst",
    "34.": "vierunddreißigst",
    "35.": "fünfunddreißigst",
    "36.": "sechsunddreißigst",
    "37.": "siebenunddreißigst",
    "38.": "achtunddreißigst",
    "39.": "neununddreißigst",
    "40.": "vierzigst",
    "41.": "einundvierzigst",
    "42.": "zweiundvierzigst",
    "43.": "dreiundvierzigst",
    "44.": "vierundvierzigst",
    "45.": "fünfundvierzigst",
    "46.": "sechsundvierzigst",
    "47.": "siebenundvierzigst",
    "48.": "achtundvierzigst",
    "49.": "neunundvierzigst",
    "50.": "fünfzigst",
    "51.": "einundfünfzigst",
    "52.": "zweiundfünfzigst",
    "53.": "dreiundfünfzigst",
    "60.": "sechzigst",
    "70.": "siebzigst",
    "80.": "achtzigst",
    "90.": "neunzigst",
    "100.": "einhundertst",
    "1000.": "eintausendst",
    "1000000.": "millionst"
}

_LONG_SCALE = OrderedDict([
    (100, 'hundert'),
    (1000, 'tausend'),
    (1000000, 'million'),
    (1e9, "milliarde"),
    (1e12, 'billion'),
    (1e15, "billiarde"),
    (1e18, "trillion"),
    (1e21, "trilliarde"),
    (1e24, "quadrillion"),
    (1e27, "quadrilliarde")
])

_MULTIPLIER = set(_LONG_SCALE.values())

_STRING_LONG_SCALE = invert_dict(_LONG_SCALE)

# ending manipulation
for number, item in _LONG_SCALE.items():
    if int(number) > 1000:
        if item.endswith('e'):
            name = item + 'n'
            _MULTIPLIER.add(name)
            _STRING_LONG_SCALE[name] = number
        else:
            name = item + 'en'
            _MULTIPLIER.add(name)
            _STRING_LONG_SCALE[name] = number

_LONG_ORDINAL = {
    1e6: "millionst",
    1e9: "milliardst",
    1e12: "billionst",
    1e15: "billiardst",
    1e18: "trillionst",
    1e21: "trilliardst",
    1e24: "quadrillionst",
    1e27: "quadrilliardst"
}

_LONG_ORDINAL.update(_ORDINAL_BASE)

# dict für erste, drittem, millionstes ...
_STRING_LONG_ORDINAL = {ord + ending: num for ord, num in invert_dict(_LONG_ORDINAL).items()
                        for ending in ("en", "em", "es", "er", "e")}

_FRACTION_MARKER = set()

_NEGATIVES = {"minus"}

_NUMBER_CONNECTORS = {"und"}

_COMMA = {"komma", "comma", "punkt"}


def nice_number_de(number, speech=True, denominators=range(1, 21)):
    """ German helper for nice_number
    This function formats a float to human understandable functions. Like
    4.5 becomes "4 einhalb" for speech and "4 1/2" for text
    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3)).replace(".", ",")
    whole, num, den = result
    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)
    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING[den]
    if whole == 0:
        if num == 1:
            return_string = 'ein {}'.format(den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1:
        return_string = '{} und ein {}'.format(whole, den_str)
    else:
        return_string = '{} und {} {}'.format(whole, num, den_str)

    return return_string


def pronounce_number_de(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """

    # TODO short_scale, scientific and ordinals
    # currently ignored

    def pronounce_triplet_de(num):
        result = ""
        num = floor(num)
        if num > 99:
            hundreds = floor(num / 100)
            if hundreds > 0:
                number = _NUM_STRING[hundreds] if hundreds > 1 else "ein"
                result += number + 'hundert' + _EXTRA_SPACE
                num -= hundreds * 100
        if num == 0:
            result += ''  # do nothing
        elif num == 1:
            result += 'eins'  # need the s for the last digit
        elif num <= 20:
            result += _NUM_STRING[num]  # + _EXTRA_SPACE_DA
        elif num > 20:
            ones = num % 10
            tens = num - ones
            if ones > 0:
                number = _NUM_STRING[ones]
                if ones == 1 and tens > 0:  # eins > ein
                    number = number[:-1]
                result += number + _EXTRA_SPACE
                if tens > 0:
                    result += 'und' + _EXTRA_SPACE
            if tens > 0:
                result += _NUM_STRING[tens] + _EXTRA_SPACE
        return result

    def pronounce_fractional_de(num,
                                places):  # fixed number of places even with
        # trailing zeros
        result = ""
        place = 10
        while places > 0:  # doesn't work with 1.0001 and places = 2: int(
            # number*place) % 10 > 0 and places > 0:
            result += " " + _NUM_STRING[int(num * place) % 10]
            if int(num * place) % 10 == 1:
                result += 's'  # "1" is pronounced "eins" after the decimal
                # point
            place *= 10
            places -= 1
        return result

    def pronounce_whole_number_de(num, scale_level=0):
        if num == 0:
            return ''

        num = floor(num)
        result = ''
        last_triplet = num % 1000

        if last_triplet == 1:
            if scale_level == 0:
                if result != '':
                    result += '' + 'eins'
                else:
                    result += "eins"
            elif scale_level == 1:
                result += 'ein' + _EXTRA_SPACE + 'tausend' + _EXTRA_SPACE
            else:
                result += "eine " + _NUM_POWERS_OF_TEN[scale_level] + ' '
        elif last_triplet > 1:
            result += pronounce_triplet_de(last_triplet)
            if scale_level == 1:
                # result += _EXTRA_SPACE_DA
                result += 'tausend' + _EXTRA_SPACE
            if scale_level >= 2:
                # if _EXTRA_SPACE_DA == '':
                #    result += " "
                result += " " + _NUM_POWERS_OF_TEN[scale_level]
            if scale_level >= 2:
                if scale_level % 2 == 0:
                    result += "e"  # MillionE
                result += "n "  # MilliardeN, MillioneN

        num = floor(num / 1000)
        scale_level += 1
        return pronounce_whole_number_de(num,
                                         scale_level) + result  # + _EXTRA_SPACE_DA

    result = ""
    if abs(number) >= 1000000000000000000000000:  # cannot do more than this
        return str(number)
    elif number == 0:
        return str(_NUM_STRING[0])
    elif number < 0:
        return "minus " + pronounce_number_de(abs(number), places)
    else:
        if number == int(number):
            return pronounce_whole_number_de(number)
        else:
            whole_number_part = floor(number)
            fractional_part = number - whole_number_part
            result += pronounce_whole_number_de(whole_number_part)
            if places > 0:
                result += " Komma"
                result += pronounce_fractional_de(fractional_part, places)
            return result


def pronounce_ordinal_de(number):
    """
      This function pronounces a number as an ordinal

      1 -> first
      2 -> second

      Args:
          number (int): the number to format
      Returns:
          (str): The pronounced number string.
      """
    # ordinals for 1, 3, 7 and 8 are irregular
    # this produces the base form, it will have to be adapted for genus,
    # casus, numerus

    ordinals = ["nullte", "erste", "zweite", "dritte", "vierte", "fünfte",
                "sechste", "siebte", "achte"]

    # only for whole positive numbers including zero
    if number < 0 or number != int(number):
        return number
    elif number < 9:
        return ordinals[number]
    elif number < 20:
        return pronounce_number_de(number) + "te"
    else:
        return pronounce_number_de(number) + "ste"


def numbers_to_digits_de(text, short_scale=False,
                                 ordinals=False, fractions=True):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numberres should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)
    Returns:
        str
        The original text, with numbers subbed in where appropriate.
    """
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_de(tokens, short_scale, ordinals, fractions)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_de(tokens, short_scale=True,
                                  ordinals=False, fractions=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_de(tokens, short_scale,
                                         ordinals)

        if not to_replace:
            break

        if isinstance(to_replace.value, float) and not fractions:
            pass
        else:
            results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_de(tokens, short_scale=True,
                                 ordinals=False):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = \
        _extract_number_with_text_de_helper(tokens, short_scale,
                                            ordinals)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_de_helper(tokens,
                                        short_scale, ordinals):
    """
    Helper for _extract_number_with_text_de.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:
    Returns:
        int or float, [Tokens]
    """
    if ordinals:
        for token in tokens:
            ordinal = is_ordinal_de(token.word)
            if ordinal:
                return ordinal, [token]

    return _extract_real_number_with_text_de(tokens, short_scale)


def _extract_real_number_with_text_de(tokens, short_scale):
    """
    This is handling real numbers.

    Args:
        tokens [Token]:
        short_scale boolean:
    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.
    """
    number_words = []
    val = _val = _current_val = None
    _comma = False
    to_sum = []

    for idx, token in enumerate(tokens):

        _prev_val = _current_val
        _current_val = None

        word = token.word

        if word in _NUMBER_CONNECTORS and not number_words:
            continue
        if word in (_NEGATIVES | _NUMBER_CONNECTORS | _COMMA):
            number_words.append(token)
            if word in _COMMA:
                _comma = token
                _current_val = _val or _prev_val
            continue

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""

        if word not in _STRING_LONG_SCALE and \
                word not in _STRING_NUM and \
                word not in _MULTIPLIER and \
                not is_numeric_de(word) and \
                not is_fractional_de(word):
            words_only = [token.word for token in number_words]
            if _val is not None:
                to_sum.append(_val)
            if to_sum:
                val = sum(to_sum)

            if number_words and (not all([w in _ARTICLES | _NEGATIVES
                                          | _NUMBER_CONNECTORS for w in words_only])
                                 or str(val) == number_words[-1].word):
                break
            else:
                number_words.clear()
                to_sum.clear()
                val = _val = _prev_val = None
            continue
        elif word not in _MULTIPLIER \
                and prev_word not in _MULTIPLIER \
                and prev_word not in _NUMBER_CONNECTORS \
                and prev_word not in _NEGATIVES \
                and prev_word not in _COMMA \
                and prev_word not in _STRING_LONG_SCALE \
                and prev_word not in _STRING_NUM \
                and not is_ordinal_de(word) \
                and not is_numeric_de(prev_word) \
                and not is_fractional_de(prev_word):
            number_words = [token]
        else:
            number_words.append(token)

        # is this word already a number or a word of a number?
        _val = _current_val = is_number_de(word)

        # is this a negative number?
        if _current_val is not None and prev_word in _NEGATIVES:
            _val = 0 - _current_val

        # is the prev word a number and should we multiply it?
        if _prev_val is not None and (word in _MULTIPLIER or \
                                      word in ("einer", "eines", "einem")):
            to_sum.append(_prev_val * _current_val or _current_val)
            _val = _current_val = None

        # fraction handling
        _fraction_val = is_fractional_de(word, short_scale=short_scale)
        if _fraction_val:
            if _prev_val is not None and prev_word != "eine" and \
                    word not in _STRING_FRACTION:  # zusammengesetzter Bruch
                _val = _prev_val + _fraction_val
                if prev_word not in _NUMBER_CONNECTORS and tokens[idx - 1] not in number_words:
                    number_words.append(tokens[idx - 1])
            elif _prev_val is not None:
                _val = _prev_val * _fraction_val
                if tokens[idx - 1] not in number_words:
                    number_words.append(tokens[idx - 1])
            else:
                _val = _fraction_val
            _current_val = _val

        # directly following numbers without relation
        if (is_numeric_de(prev_word) or prev_word in _STRING_NUM) \
                and not _fraction_val and not is_fractional_de(next_word) and not to_sum:
            val = _prev_val
            number_words.pop(-1)
            break

        # is this a spoken time ("drei viertel acht")
        if isinstance(_prev_val, float) and is_number_de(word) and not to_sum:
            if idx + 1 < len(tokens):
                _, number = _extract_real_number_with_text_de([tokens[idx + 1]],
                                                              short_scale=short_scale)
            if not next_word or not number:
                val = f"{_val - 1}:{int(60 * _prev_val)}"
                break

        # correct time format (whisper "13.30 Uhr")
        if all([isinstance(_current_val, float),
                next_word.lower() in ["uhr", "pm", "a.m.", "p.m."]]):
            components = word.split(".")
            if len(components) == 2 and \
                    all(map(str.isdigit, components)) and \
                    int(components[0]) < 25 and int(components[1]) < 60:
                _hstr, _mstr = components
                _mstr = _mstr.ljust(2, "0")
                tokens[idx] = Token(f"{_hstr}:{_mstr}", idx)
                number_words.clear()
                _val = _prev_val = None
                continue

                # spoken decimals
        if _current_val is not None and _comma:
            # to_sum = [ 1, 0.2, 0.04,...]
            to_sum.append(_current_val if _current_val >= 10 else (
                                                                      _current_val) / (
                                                                          10 ** (token.index - _comma.index)))
            _val = _current_val = None

        if _current_val is not None and next_word in (_NUMBER_CONNECTORS | _COMMA | {""}):
            to_sum.append(_val or _current_val)
            _val = _current_val = None

        if not next_word and number_words:
            val = sum(to_sum) or _val

    return val, number_words


def is_fractional_de(input_str, short_scale=False):
    """
    This function takes the given text and checks if it is a fraction.
    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction
    """
    # account for different numerators, e.g. zweidrittel

    input_str = input_str.lower()
    numerator = 1
    prev_number = 0
    denominator = False
    remainder = ""

    # first check if is a fraction containing a char (eg "2/3")
    _bucket = input_str.split('/')
    if look_for_fractions(_bucket):
        numerator = float(_bucket[0])
        denominator = float(_bucket[1])

    if not denominator:
        for fraction in sorted(_STRING_FRACTION.keys(),
                               key=lambda x: len(x),
                               reverse=True):
            if fraction in input_str and not denominator:
                denominator = _STRING_FRACTION.get(fraction)
                remainder = input_str.replace(fraction, "")
                break

        if remainder:
            if not _STRING_NUM.get(remainder, False):
                # acount for eineindrittel
                for numstring, number in _STRING_NUM.items():
                    if remainder.endswith(numstring):
                        prev_number = _STRING_NUM.get(
                            remainder.replace(numstring, "", 1), 0)
                        numerator = number
                        break
                else:
                    return False
            else:
                numerator = _STRING_NUM.get(remainder)

    if denominator:
        return prev_number + (numerator / denominator)
    else:
        return False


def is_ordinal_de(input_str):
    """
    This function takes the given text and checks if it is an ordinal number.
    Args:
        input_str (str): the string to check if ordinal
    Returns:
        (bool) or (float): False if not an ordinal, otherwise the number
        corresponding to the ordinal
    ordinals for 1, 3, 7 and 8 are irregular
    only works for ordinals corresponding to the numbers in _STRING_NUM
    """
    val = _STRING_LONG_ORDINAL.get(input_str.lower(), False)
    # account for numbered ordinals
    if not val and input_str.endswith('.') and is_numeric_de(input_str[:-1]):
        val = input_str
    return val


def _get_ordinal_index(input_str: str, type_: type = str):
    ord = is_ordinal_de(input_str)
    return type_(ord.replace(".", "")) if ord else ord


def is_number_de(word: str):
    if is_numeric_de(word):
        if word.isdigit():
            return int(word)
        else:
            return float(word)
    elif word in _STRING_NUM:
        return _STRING_NUM.get(word)
    elif word in _STRING_LONG_SCALE:
        return _STRING_LONG_SCALE.get(word)

    return None


def is_numeric_de(input_str):
    """
    Takes in a string and tests to see if it is a number.

    Args:
        text (str): string to test if a number
    Returns:
        (bool): True if a number, else False
    """
    # da float("1.") = 1.0
    if input_str.endswith('.'):
        return False
    try:
        float(input_str)
        return True
    except ValueError:
        return False


def extract_number_de(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    numbers = _extract_numbers_with_text_de(tokenize(text.lower()),
                                            short_scale, ordinals)
    # if query ordinals only consider ordinals
    if ordinals:
        numbers = list(filter(lambda x: isinstance(x.value, str)
                                        and x.value.endswith("."),
                              numbers))

    number = numbers[0].value if numbers else None

    if number:
        number = float(number)
        if number.is_integer():
            number = int(number)

    return number



from collections import namedtuple
from enum import Enum
from quebra_frases import word_tokenize

# Token is intended to be used in the number processing functions in
# this module. The parsing requires slicing and dividing of the original
# text. To ensure things parse correctly, we need to know where text came
# from in the original input, hence this nametuple.
Token = namedtuple('Token', 'word index')


class Scale(str, Enum):
    """
    Defines the numerical scale to be used.
    - SHORT: Short scale (e.g., billion = 10^9).
    - LONG: Long scale (e.g., billion = 10^12).
    """
    SHORT = "short"
    LONG = "long"


class GrammaticalGender(str, Enum):
    """
    Defines the grammatical gender for number pronunciation.
    """
    MASCULINE = "masculine"
    FEMININE = "feminine"
    NEUTRAL = "neutral"


class DigitPronunciation(str, Enum):
    DIGIT_BY_DIGIT = "digit"
    FULL_NUMBER = "number"


class ReplaceableNumber:
    """
    Similar to Token, this class is used in number parsing.

    Once we've found a number in a string, this class contains all
    the info about the value, and where it came from in the original text.
    In other words, it is the text, and the number that can replace it in
    the string.
    """

    def __init__(self, value, tokens: [Token]):
        self.value = value
        self.tokens = tokens

    def __bool__(self):
        return bool(self.value is not None and self.value is not False)

    @property
    def start_index(self):
        return self.tokens[0].index

    @property
    def end_index(self):
        return self.tokens[-1].index

    @property
    def text(self):
        """
        Return the concatenated text represented by the tokens, separated by spaces.
        """
        return ' '.join([str(t.word) for t in self.tokens if t.word])

    def __setattr__(self, key, value):
        """
        Prevent modification of existing attributes, allowing only new attributes to be set.

        Raises:
            Exception: If attempting to modify an attribute that already exists.
        """
        try:
            getattr(self, key)
        except AttributeError:
            super().__setattr__(key, value)
        else:
            raise Exception("Immutable!")

    def __str__(self):
        return "({v}, {t})".format(v=self.value, t=self.tokens)

    def __repr__(self):
        return "{n}({v}, {t})".format(n=self.__class__.__name__, v=self.value,
                                      t=self.tokens)


def tokenize(text):
    """
    Generate a list of token object, given a string.
    Args:
        text str: Text to tokenize.

    Returns:
        [Token]

    """
    return [Token(word, index)
            for index, word in enumerate(word_tokenize(text))]


def partition_list(items, split_on):
    """
    Partition a list of items.

    Works similarly to str.partition

    Args:
        items:
        split_on callable:
            Should return a boolean. Each item will be passed to
            this callable in succession, and partitions will be
            created any time it returns True.

    Returns:
        [[any]]

    """
    splits = []
    current_split = []
    for item in items:
        if split_on(item):
            splits.append(current_split)
            splits.append([item])
            current_split = []
        else:
            current_split.append(item)
    splits.append(current_split)
    return list(filter(lambda x: len(x) != 0, splits))


def invert_dict(original):
    """
    Produce a dictionary with the keys and values
    inverted, relative to the dict passed in.

    Args:
        original dict: The dict like object to invert

    Returns:
        dict

    """
    return {value: key for key, value in original.items()}


def is_numeric(input_str):
    """
    Return True if the input string represents a valid number, otherwise False.

    Parameters:
        input_str (str): The string to test for numeric value.

    Returns:
        bool: True if the string can be converted to a float, False otherwise.
    """
    try:
        float(input_str)
        return True
    except ValueError:
        return False


def look_for_fractions(split_list):
    """"
    This function takes a list made by fraction & determines if a fraction.

    Args:
        split_list (list): list created by splitting on '/'
    Returns:
        (bool): False if not a fraction, otherwise True

    """

    if len(split_list) == 2:
        if is_numeric(split_list[0]) and is_numeric(split_list[1]):
            return True

    return False


def convert_to_mixed_fraction(number, denominators=range(1, 21)):
    """
    Convert floats to components of a mixed fraction representation

    Returns the closest fractional representation using the
    provided denominators.  For example, 4.500002 would become
    the whole number 4, the numerator 1 and the denominator 2

    Args:
        number (float): number for convert
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        whole, numerator, denominator (int): Integers of the mixed fraction
    """
    int_number = int(number)
    if int_number == number:
        return int_number, 0, 1  # whole number, no fraction

    frac_number = abs(number - int_number)
    if not denominators:
        denominators = range(1, 21)

    for denominator in denominators:
        numerator = abs(frac_number) * denominator
        if abs(numerator - round(numerator)) < 0.01:  # 0.01 accuracy
            break
    else:
        return None

    return int_number, int(round(numerator)), denominator
